#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from piquasso.api.result import Result

from piquasso.api.program import Program


class Computer(abc.ABC):
    """Base class for all quantum computers or simulators supported by Piquasso."""

    @abc.abstractmethod
    def execute(
        self,
        program: Program,
        shots: int = 1,
    ) -> Result:
        """Executes the program `shots` times.

        Args:
            program (Program): The program to be executes.
            shots (int, optional):
                The number of times the program should be executed. Defaults to 1.

        Returns:
            Result: The result of the execution containing `shots` number of samples.
        """
        pass



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple

import numpy as np

from scipy.special import factorial

from piquasso._math.linalg import block_reduce, reduce_
from piquasso._math.torontonian import torontonian

from piquasso.api.typing import HafnianFunction


class DensityMatrixCalculation:
    def __init__(
        self,
        complex_displacement: np.ndarray,
        complex_covariance: np.ndarray,
        loop_hafnian_function: HafnianFunction,
    ) -> None:
        d = len(complex_displacement) // 2
        Q = (complex_covariance + np.identity(2 * d)) / 2

        Qinv = np.linalg.inv(Q)
        identity = np.identity(d)
        zeros = np.zeros_like(identity)

        X = np.block(
            [
                [zeros, identity],
                [identity, zeros],
            ],
        )

        self._A: np.ndarray = X @ (np.identity(2 * d, dtype=complex) - Qinv)

        self._gamma: np.ndarray = complex_displacement.conj() @ Qinv

        self._normalization: np.ndarray = np.exp(
            -0.5 * self._gamma @ complex_displacement
        ) / np.sqrt(np.linalg.det(Q))

        self.loop_hafnian_function = loop_hafnian_function

    def _get_A_reduced(self, reduce_on: Tuple[int, ...]) -> np.ndarray:
        A_reduced = reduce_(self._A, reduce_on=reduce_on)

        np.fill_diagonal(A_reduced, reduce_(self._gamma, reduce_on=reduce_on))

        return A_reduced

    def get_density_matrix_element(
        self, bra: Tuple[int, ...], ket: Tuple[int, ...]
    ) -> float:
        reduce_on = ket + bra

        A_reduced = self._get_A_reduced(reduce_on=reduce_on)

        return (
            self._normalization
            * self.loop_hafnian_function(A_reduced)
            / np.sqrt(np.prod(factorial(reduce_on)))
        )

    def get_density_matrix(
        self, occupation_numbers: List[Tuple[int, ...]]
    ) -> np.ndarray:
        n = len(occupation_numbers)

        density_matrix = np.empty(shape=(n, n), dtype=complex)

        for i, bra in enumerate(occupation_numbers):
            for j, ket in enumerate(occupation_numbers):
                density_matrix[i, j] = self.get_density_matrix_element(bra, ket)

        return density_matrix

    def get_particle_number_detection_probabilities(
        self, occupation_numbers: List[Tuple[int, ...]]
    ) -> np.ndarray:
        ret_list = []

        for occupation_number in occupation_numbers:
            probability = np.real(
                self.get_density_matrix_element(
                    bra=occupation_number,
                    ket=occupation_number,
                )
            )
            ret_list.append(probability)

        ret = np.array(ret_list, dtype=float)

        ret[abs(ret) < 1e-10] = 0.0

        return ret


def calculate_click_probability(
    xp_covariance: np.ndarray,
    occupation_number: Tuple[int, ...],
    hbar: float,
) -> float:
    r"""
    Calculates the threshold detection probability with the equation

    .. math::
        p(S) = \frac{
            \operatorname{tor}( I - ( \Sigma^{-1} )_{(S)} )
        }{
            \sqrt{\operatorname{det}(\Sigma)}
        },

    where :math:`\Sigma \in \mathbb{R}^{2d \times 2d}` is a symmetric matrix
    defined by

    .. math::
        \Sigma = \frac{1}{2} \left (
                \frac{1}{\hbar} \sigma_{xp}
                + I
            \right ).
    """

    d = len(xp_covariance) // 2

    sigma: np.ndarray = (xp_covariance / hbar + np.identity(2 * d)) / 2

    sigma_inv_reduced = block_reduce(
        np.linalg.inv(sigma),
        reduce_on=occupation_number,
    )

    probability = (
        torontonian(
            np.identity(len(sigma_inv_reduced), dtype=float) - sigma_inv_reduced
        )
    ).real / np.sqrt(np.linalg.det(sigma))

    return max(probability, 0.0)



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class PiquassoException(Exception):
    """Base class for all exceptions raised by Piquasso."""


class InvalidState(PiquassoException):
    """Raised when an invalid state is encountered or being prepared."""


class InvalidParameter(PiquassoException):
    """Raised when an invalid parameter is specified."""


class InvalidModes(PiquassoException):
    """Raised when invalid set of modes are encountered."""


class InvalidProgram(PiquassoException):
    """Raised when an invalid program is being created or used."""


class InvalidInstruction(PiquassoException):
    """Raised when an invalid instruction is specified."""


class InvalidSimulation(PiquassoException):
    """Raised when a simulation could not be executed."""



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piquasso.api.simulator import Simulator
from piquasso.instructions import preparations, gates, measurements, channels

from .state import SamplingState
from .calculations import state_vector, passive_linear, sampling, loss


class SamplingSimulator(Simulator):
    """Performs photonic simulations using Fock representation with pure states.

    The simulation (when executed) results in an instance of
    :class:`~piquasso._backends.sampling.state.SamplingState`.

    Example usage::

        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

        simulator = pq.SamplingSimulator(d=5)
        result = simulator.execute(program)

    Supported preparations:
        :class:`~piquasso.instructions.preparations.StateVector`.

    Supported gates:
        :class:`~piquasso.instructions.gates.Interferometer`,
        :class:`~piquasso.instructions.gates.Beamsplitter`,
        :class:`~piquasso.instructions.gates.Phaseshifter`,
        :class:`~piquasso.instructions.gates.MachZehnder`,
        :class:`~piquasso.instructions.gates.Fourier`,
        :class:`~piquasso.instructions.gates.Interferometer`.

    Supported measurements:
        :class:`~piquasso.instructions.measurements.Sampling`.

    Supported channels:
        :class:`~piquasso.instructions.channels.Loss`.
    """

    _instruction_map = {
        preparations.StateVector: state_vector,
        gates.Beamsplitter: passive_linear,
        gates.Phaseshifter: passive_linear,
        gates.MachZehnder: passive_linear,
        gates.Fourier: passive_linear,
        gates.Interferometer: passive_linear,
        measurements.Sampling: sampling,
        channels.Loss: loss,
    }

    _state_class = SamplingState



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import copy
import typing
from typing import TypeVar, Generic

if typing.TYPE_CHECKING:
    from piquasso.api.program import Program
    from piquasso.api.mode import Q

_T = TypeVar("_T", float, complex)


class DictMixin(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def from_dict(cls, dict_: dict) -> "DictMixin":
        """Creates an instance from a `dict` specified.

        Args:
            dict_ (dict): The desired instance in the format of a `dict`.
        """
        pass


class WeightMixin(abc.ABC, Generic[_T]):
    @property
    @abc.abstractmethod
    def params(self) -> dict:
        pass

    def __mul__(self, coefficient: _T) -> "WeightMixin":
        self.params["coefficient"] *= coefficient
        return self

    __rmul__ = __mul__

    def __truediv__(self, coefficient: _T) -> "WeightMixin":
        return self.__mul__(1 / coefficient)


class RegisterMixin(abc.ABC):
    @abc.abstractmethod
    def _apply_to_program_on_register(self, program: "Program", register: "Q") -> None:
        """Applies the current object to the specifed program on its specified register.

        Args:
            program (Program): [description]
            register (Q): [description]
        """
        pass

    def copy(self) -> "RegisterMixin":
        """Copies the current object with :func:`copy.deepcopy`.

        Returns:
            A deepcopy of the current object.
        """
        return copy.deepcopy(self)


class CodeMixin(abc.ABC):
    @abc.abstractmethod
    def _as_code(self) -> str:
        pass


class ScalingMixin(abc.ABC):
    @abc.abstractmethod
    def _autoscale(self) -> None:
        pass



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain, combinations, combinations_with_replacement
from typing import Tuple, Iterable, Iterator, TypeVar, List, Type

import numpy as np

_T = TypeVar("_T")


def powerset(iterable: Iterable[_T]) -> Iterator[Tuple[_T, ...]]:
    s = list(iterable)
    return chain.from_iterable(combinations(iterable, r) for r in range(len(s) + 1))


def partitions(
    boxes: int, particles: int, class_: Type[tuple] = tuple
) -> List[Tuple[int, ...]]:
    if particles == 0:
        return [class_([0] * boxes)]

    masks = np.rot90(np.identity(boxes, dtype=int))

    return sorted(
        class_(sum(c))  # type: ignore
        for c in combinations_with_replacement(masks, particles)
    )


def get_occupation_numbers(d: int, cutoff: int) -> List[Tuple[int, ...]]:
    return [
        occupation_number
        for particle_number in range(cutoff)
        for occupation_number in partitions(d, particle_number)
    ]



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from piquasso._math.linalg import is_positive_semidefinite, is_real_2n_by_2n
from piquasso._math.symplectic import symplectic_form

from piquasso.core import _mixins

from piquasso.api.errors import InvalidParameter
from piquasso.api.instruction import Gate


class DeterministicGaussianChannel(Gate):
    r"""Deterministic Gaussian channel.

    It is a CP (completely positive) map between Gaussian states which is characterized
    by the matrices :math:`X, Y \in \mathbb{R}^{2n \times 2n}` acting on the mean vector
    and covariance matrix with the mapping

    .. math::
        \mu &\mapsto X \mu \\
        \sigma &\mapsto X \sigma X^T + Y.

    The matrices :math:`X` and :math:`Y` should satisfy the inequality

    .. math::
        Y + i \Omega \geq i X \Omega X^T.

    Note:
        The matrix :math:`Y` is dependent on :math:`\hbar`, but the value of
        :math:`\hbar` is specified later when executed by a simulator. The parameter
        `Y` should be specified keeping in mind that it will automatically be scaled
        with :math:`\hbar` during execution.

    Note:
        Currently, this instruction can only be used along with
        :class:`~piquasso._backends.gaussian.state.GaussianState`.
    """

    def __init__(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Args:
            X (numpy.ndarray):
                Transformation matrix on the quadrature vectors in `xpxp` order.
            Y (numpy.ndarray):
                The additive noise contributing to the covariance matrix in `xpxp`
                order.

        Raises:
            InvalidParameter: If the specified 'X' and/or 'Y' matrices are invalid.
        """

        if not is_real_2n_by_2n(X):
            raise InvalidParameter(
                f"The parameter 'X' must be a real 2n-by-2n matrix: X={X}"
            )

        if not is_real_2n_by_2n(Y):
            raise InvalidParameter(
                f"The parameter 'Y' must be a real 2n-by-2n matrix: Y={Y}"
            )

        if X.shape != Y.shape:
            raise InvalidParameter(
                f"The shape of matrices 'X' and 'Y' should be equal: "
                f"X.shape={X.shape}, Y.shape={Y.shape}"
            )

        omega = symplectic_form(len(X) // 2)

        if not is_positive_semidefinite(
            Y - 1j * omega - 1j * X @ omega @ X.transpose()
        ):
            raise InvalidParameter(
                "The matrices 'X' and 'Y' does not satisfy the inequality "
                "corresponding to Gaussian channels."
            )

        super().__init__(params=dict(X=X, Y=Y))


class Attenuator(Gate):
    r"""The attenuator channel.

    It reduces the first moments' amplitude by :math:`\cos \theta` by mixing the input
    state with a thermal state using a beamsplitter.

    It can also be characterized as a deterministic Gaussian channel with mappings

    .. math::
        X &= \cos \theta I_{2 \times 2} \\
        Y &= ( \sin \theta )^2 (2 \bar{N} + 1) I_{2 \times 2},

    where :math:`\theta \in [0, 2 \pi)` is the beampsplitters' mixing angle and
    :math:`\bar{N} \in \mathbb{R}^{+}`. :math:`\bar{N}` is the mean number of thermal
    excitations of the system interacting with the environment.

    Note:
        Currently, this instruction can only be used along with
        :class:`~piquasso._backends.gaussian.state.GaussianState`.
    """

    def __init__(self, theta: float, mean_thermal_excitation: float = 0) -> None:
        """
        Args:
            theta (float): The mixing angle.
            mean_thermal_excitation (int):
                Mean number of thermal excitations of the system interacting with the
                environment.

        Raises:
            InvalidParameter: If the specified mean thermal excitation is not positive.
        """

        if mean_thermal_excitation < 0:
            raise InvalidParameter(
                "The parameter 'mean_thermal_excitation' must be a positive real "
                f"number: mean_thermal_excitation={mean_thermal_excitation}"
            )

        X = np.cos(theta) * np.identity(2)

        Y = (np.sin(theta) ** 2) * (2 * mean_thermal_excitation + 1) * np.identity(2)

        super().__init__(
            params=dict(theta=theta, mean_thermal_excitation=mean_thermal_excitation),
            extra_params=dict(X=X, Y=Y),
        )


class Loss(Gate, _mixins.ScalingMixin):
    """Applies a loss channel to the state.

    Note:
        Currently, this instruction can only be used along with
        :class:`~piquasso._backends.sampling.state.SamplingState`.
    """

    def __init__(self, transmissivity: np.ndarray) -> None:
        """
        Args:
            transmissivity (numpy.ndarray): The transmissivity array.
        """
        super().__init__(
            params=dict(transmissivity=transmissivity),
            extra_params=dict(
                transmissivity=np.atleast_1d(transmissivity),
            ),
        )

    def _autoscale(self) -> None:
        transmissivity = self._extra_params["transmissivity"]
        if transmissivity is None or len(self.modes) == len(transmissivity):
            pass
        elif len(transmissivity) == 1:
            self._extra_params["transmissivity"] = np.array(
                [transmissivity[0]] * len(self.modes),
                dtype=complex,
            )
        else:
            raise InvalidParameter(
                f"The channel {self} is not applicable to modes {self.modes} with the "
                "specified parameters."
            )



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Generator, Any, Dict

import numpy as np

from piquasso.api.config import Config
from piquasso.api.errors import InvalidState, PiquassoException
from piquasso._math.fock import cutoff_cardinality, FockBasis

from ..state import BaseFockState
from ..general.state import FockState


class PureFockState(BaseFockState):
    r"""A simulated pure Fock state.

    If no mixed states are needed for a Fock simulation, then this state is the most
    appropriate currently, since it does not store an entire density matrix, only a
    state vector with size

    .. math::
        {d + c - 1 \choose c - 1},

    where :math:`c \in \mathbb{N}` is the Fock space cutoff.
    """

    def __init__(self, *, d: int, config: Config = None) -> None:
        """
        Args:
            state_vector (numpy.ndarray, optional): The initial state vector.
            d (int): The number of modes.
            cutoff (int): The Fock space cutoff.
        """

        super().__init__(d=d, config=config)

        self._state_vector = self._get_empty()

    def _get_empty(self) -> np.ndarray:
        return np.zeros(shape=(self._space.cardinality,), dtype=complex)

    def reset(self) -> None:
        self._state_vector = self._get_empty()
        self._state_vector[0] = 1.0

    @property
    def nonzero_elements(self) -> Generator[Tuple[complex, FockBasis], Any, None]:
        for index, basis in self._space.basis:
            coefficient: complex = self._state_vector[index]
            if coefficient != 0:
                yield coefficient, basis

    def __repr__(self) -> str:
        return " + ".join(
            [
                str(coefficient) + str(basis)
                for coefficient, basis in self.nonzero_elements
            ]
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PureFockState):
            return False
        return np.allclose(self._state_vector, other._state_vector)

    @property
    def density_matrix(self) -> np.ndarray:
        cardinality = cutoff_cardinality(d=self.d, cutoff=self._config.cutoff)

        state_vector = self._state_vector[:cardinality]

        return np.outer(state_vector, state_vector.conj())

    def _as_mixed(self) -> FockState:
        return FockState.from_fock_state(self)

    def reduced(self, modes: Tuple[int, ...]) -> FockState:
        return self._as_mixed().reduced(modes)

    def get_particle_detection_probability(
        self, occupation_number: Tuple[int, ...]
    ) -> float:
        if len(occupation_number) != self.d:
            raise PiquassoException(
                f"The specified occupation number should have length '{self.d}': "
                f"occupation_number='{occupation_number}'."
            )

        index = self._space.index(occupation_number)

        return np.real(
            self._state_vector[index].conjugate() * self._state_vector[index]
        )

    @property
    def fock_probabilities(self) -> np.ndarray:
        cardinality = cutoff_cardinality(d=self._space.d, cutoff=self._config.cutoff)

        return (self._state_vector * self._state_vector.conjugate()).real[:cardinality]

    @property
    def fock_probabilities_map(self) -> Dict[Tuple[int, ...], float]:
        probability_map: Dict[Tuple[int, ...], float] = {}

        for index, basis in self._space.basis:
            probability_map[tuple(basis)] = np.abs(self._state_vector[index]) ** 2

        return probability_map

    def normalize(self) -> None:
        if np.isclose(self.norm, 0):
            raise InvalidState("The norm of the state is 0.")

        self._state_vector = self._state_vector / np.sqrt(self.norm)

    def validate(self) -> None:
        """Validates the represented state.

        Raises:
            InvalidState:
                Raised, if the norm of the state vector is not close to 1.0.
        """
        sum_of_probabilities = sum(self.fock_probabilities)

        if not np.isclose(sum_of_probabilities, 1.0):
            raise InvalidState(
                f"The sum of probabilities is {sum_of_probabilities}, "
                "instead of 1.0:\n"
                f"fock_probabilities={self.fock_probabilities}"
            )

    def quadratures_mean_variance(
        self, modes: Tuple[int, ...], phi: float = 0
    ) -> Tuple[float, float]:
        r"""This method calculates the mean and the variance of the qudrature operators
        for a single qumode state.
        The quadrature operators :math:`x` and :math:`p` for a mode :math:`i`
        can be calculated using the creation and annihilation operators as follows:

        .. math::
            x_i &= \sqrt{\frac{\hbar}{2}} (a_i + a_i^\dagger) \\
                p_i &= -i \sqrt{\frac{\hbar}{2}} (a_i - a_i^\dagger).

        Let :math:`\phi \in [ 0, 2 \pi )`, we can rotate the quadratures using the
        following transformation:

        .. math::
            Q_{i, \phi} = \cos\phi~x_i + \sin\phi~p_i.

        The expectation value :math:`\langle Q_{i, \phi}\rangle` can be calculated as:

        .. math::
            \operatorname{Tr}(\rho_i Q_{i, \phi}),

        where :math:`\rho_i` is the reduced density matrix of the mode :math:`i` and
        :math:`Q_{i, \phi}` is the rotated quadrature operator for a single mode and
        the variance is calculated as:

        .. math::
            \operatorname{\textit{Var}(Q_{i,\phi})} = \langle Q_{i, \phi}^{2}\rangle
                - \langle Q_{i, \phi}\rangle^{2}.

        Args:
            phi (float): The rotation angle. By default it is `0` which means that
                the mean of the position operator is being calculated. For :math:`\phi=
                \frac{\pi}{2}` the mean of the momentum operator is being calculated.
            modes (tuple[int]): The correspoding mode at which the mean of the
                quadratures are being calculated.
        Returns:
            (float, float): A tuple that contains the expectation value and the
                varianceof of the quadrature operator respectively.
        """
        reduced_dm = self.reduced(modes=modes).density_matrix
        annih = np.diag(np.sqrt(np.arange(1, self._config.cutoff)), 1)
        create = annih.T
        position = (create + annih) * np.sqrt(self._config.hbar / 2)
        momentum = -1j * (annih - create) * np.sqrt(self._config.hbar / 2)

        if phi != 0:
            rotated_quadratures = position * np.cos(phi) + momentum * np.sin(phi)
        else:
            rotated_quadratures = position

        expctation = np.trace(np.dot(reduced_dm, rotated_quadratures)).real
        variance = (
            np.trace(
                np.dot(reduced_dm, np.dot(rotated_quadratures, rotated_quadratures))
            ).real
            - expctation**2
        )
        return expctation, variance



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar

import numpy as np

from .combinatorics import powerset

TNum = TypeVar("TNum", np.float64, np.complex128)


def torontonian(A: np.ndarray) -> complex:
    d = A.shape[0] // 2

    if d == 0:
        return 1.0 + 0j

    ret = 0.0 + 0j

    for subset in powerset(range(0, d)):
        index = np.ix_(subset, subset)

        A_reduced = np.block(
            [
                [A[:d, :d][index], A[:d, d:][index]],
                [A[d:, :d][index], A[d:, d:][index]],
            ]
        )

        factor = 1.0 if ((d - len(subset)) % 2 == 0) else -1.0

        ret += factor / np.sqrt(
            np.linalg.det(np.identity(len(A_reduced)) - A_reduced).real + 0.0j
        )

    return ret



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from typing import Tuple, Any, Type, Dict

import numpy as np

from .mode import Q
from piquasso.core import _mixins
from piquasso.api.errors import PiquassoException

if typing.TYPE_CHECKING:
    from piquasso.api.program import Program


class Instruction(_mixins.DictMixin, _mixins.RegisterMixin, _mixins.CodeMixin):
    """
    Base class for all instructions.

    Args:
        params: Mapping of parameters specified by the users.
        extra_params: Mapping of extra parameters, typically calculated ones.
    """

    _subclasses: Dict[str, Type["Instruction"]] = {}

    def __init__(self, *, params: dict = None, extra_params: dict = None) -> None:
        self._params: dict = params or dict()

        self._extra_params: dict = extra_params or dict()

    @property
    def params(self) -> dict:
        return self._params

    @property
    def _all_params(self) -> dict:
        return {**self._params, **self._extra_params}

    def _as_code(self) -> str:
        if hasattr(self, "modes"):
            mode_string = ", ".join([str(mode) for mode in self.modes])
        else:
            mode_string = ""

        if hasattr(self, "params"):
            params_string = "{}".format(
                ", ".join(
                    [
                        f"{key}={self._param_repr(value)}"
                        for key, value in self.params.items()
                    ]
                )
            )
        else:
            params_string = ""

        return f"pq.Q({mode_string}) | pq.{self.__class__.__name__}({params_string})"

    @staticmethod
    def _param_repr(value: Any) -> str:
        if isinstance(value, np.ndarray):
            return "np." + repr(value)

        return value

    def on_modes(self, *modes: int) -> "Instruction":
        self.modes: Tuple[int, ...] = modes
        return self

    def _apply_to_program_on_register(self, program: "Program", register: Q) -> None:
        program.instructions.append(self.on_modes(*register.modes))

    @classmethod
    def from_dict(cls, dict_: dict) -> "Instruction":
        """Creates an :class:`Instruction` instance from a dict specified.

        Args:
            dict_ (dict):
                The desired :class:`Instruction` instance in the format of a `dict`.

        Returns:
            Instruction:
                An :class:`Instruction` initialized using the specified `dict`.
        """

        class_ = cls.get_subclass(dict_["type"])

        instruction = class_(**dict_["attributes"]["constructor_kwargs"])

        instruction.modes = dict_["attributes"]["modes"]

        return instruction

    @classmethod
    def set_subclass(cls, instruction: Type["Instruction"]) -> None:
        """Registers a class in the instruction subclass map.

        This is meaningful in contexts when one has multiple instructions with the same
        name.

        Example:
            When one creates a custom beamsplitter with name `Beamsplitter` and
            subclasses :class:`~piquasso.instructions.gates.Beamsplitter`, then for e.g.
            executing a Blackbird code will be performed with the custom one, not the
            original one. When one wants to use the original one in this case, one can
            reset it with this method.

        Args:
            instruction (Type[Instruction]): The instruction class to be registered.

        Raises:
            PiquassoException:
                When the class is not actually an instance of :class:`Insruction`.
        """

        if not issubclass(instruction, Instruction):
            raise PiquassoException(
                f"The instruction '{instruction}' needs to be a subclass of "
                "'pq.Instruction'."
            )

        cls._subclasses[instruction.__name__] = instruction

    @classmethod
    def get_subclass(cls, name: str) -> Type["Instruction"]:
        """Returns the instruction subclass specified by its name.

        Returns:
            Type[Instruction]: The instruction class.
        """

        return cls._subclasses[name]

    def __repr__(self) -> str:
        if hasattr(self, "modes"):
            modes = "modes={}".format(self.modes)
        else:
            modes = ""

        if getattr(self, "params") != {}:
            params = "{}, ".format(
                ", ".join([f"{key}={value}" for key, value in self.params.items()])
            )
        else:
            params = ""

        classname = self.__class__.__name__

        return f"<pq.{classname}({params}{modes})>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Instruction):
            return False
        return self.modes == other.modes and self.params == other.params

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        cls.set_subclass(cls)


class Preparation(Instruction):
    """Base class for preparations."""


class Gate(Instruction):
    """Base class for gates."""


class Measurement(Instruction):
    r"""Base class for measurements."""



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piquasso.api.simulator import Simulator
from piquasso.instructions import preparations, gates, measurements, channels

from .state import GaussianState
from .calculations import (
    passive_linear,
    linear,
    displacement,
    graph,
    homodyne_measurement,
    generaldyne_measurement,
    vacuum,
    mean,
    covariance,
    particle_number_measurement,
    threshold_measurement,
    deterministic_gaussian_channel,
)


class GaussianSimulator(Simulator):
    """Performs photonic simulations using Gaussian representation.

    The simulation (when executed) results in an instance of
    :class:`~piquasso._backends.gaussian.state.GaussianState`.

    Example usage::

        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

        simulator = pq.GaussianSimulator(d=5)
        result = simulator.execute(program)

    Supported preparations:
        :class:`~piquasso.instructions.preparations.Vacuum`,
        :class:`~piquasso.instructions.preparations.Mean`,
        :class:`~piquasso.instructions.preparations.Covariance`,
        :class:`~piquasso.instructions.preparations.Thermal`.

    Supported gates:
        :class:`~piquasso.instructions.gates.Interferometer`,
        :class:`~piquasso.instructions.gates.Beamsplitter`,
        :class:`~piquasso.instructions.gates.Phaseshifter`,
        :class:`~piquasso.instructions.gates.MachZehnder`,
        :class:`~piquasso.instructions.gates.Fourier`,
        :class:`~piquasso.instructions.gates.GaussianTransform`,
        :class:`~piquasso.instructions.gates.Squeezing`,
        :class:`~piquasso.instructions.gates.QuadraticPhase`,
        :class:`~piquasso.instructions.gates.Squeezing2`,
        :class:`~piquasso.instructions.gates.ControlledX`,
        :class:`~piquasso.instructions.gates.ControlledZ`,
        :class:`~piquasso.instructions.gates.Displacement`,
        :class:`~piquasso.instructions.gates.PositionDisplacement`,
        :class:`~piquasso.instructions.gates.MomentumDisplacement`,
        :class:`~piquasso.instructions.gates.Graph`.

    Supported measurements:
        :class:`~piquasso.instructions.measurements.HomodyneMeasurement`,
        :class:`~piquasso.instructions.measurements.HeterodyneMeasurement`,
        :class:`~piquasso.instructions.measurements.GeneraldyneMeasurement`,
        :class:`~piquasso.instructions.measurements.ParticleNumberMeasurement`,
        :class:`~piquasso.instructions.measurements.ThresholdMeasurement`.

    Supported channels:
        :class:`~piquasso.instructions.channels.DeterministicGaussianChannel`,
        :class:`~piquasso.instructions.channels.Attenuator`.
    """

    _instruction_map = {
        preparations.Vacuum: vacuum,
        preparations.Mean: mean,
        preparations.Covariance: covariance,
        preparations.Thermal: covariance,
        gates.Interferometer: passive_linear,
        gates.Beamsplitter: passive_linear,
        gates.Phaseshifter: passive_linear,
        gates.MachZehnder: passive_linear,
        gates.Fourier: passive_linear,
        gates.GaussianTransform: linear,
        gates.Squeezing: linear,
        gates.QuadraticPhase: linear,
        gates.Squeezing2: linear,
        gates.ControlledX: linear,
        gates.ControlledZ: linear,
        gates.Displacement: displacement,
        gates.PositionDisplacement: displacement,
        gates.MomentumDisplacement: displacement,
        gates.Graph: graph,
        measurements.HomodyneMeasurement: homodyne_measurement,
        measurements.HeterodyneMeasurement: generaldyne_measurement,
        measurements.GeneraldyneMeasurement: generaldyne_measurement,
        measurements.ParticleNumberMeasurement: particle_number_measurement,
        measurements.ThresholdMeasurement: threshold_measurement,
        channels.DeterministicGaussianChannel: deterministic_gaussian_channel,
        channels.Attenuator: deterministic_gaussian_channel,
    }

    _state_class = GaussianState



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import scipy
import numpy as np

from piquasso._math.symplectic import xp_symplectic_form
from piquasso._math.transformations import from_xxpp_to_xpxp_transformation_matrix
from scipy.linalg import sqrtm, schur, block_diag


def takagi(matrix, rounding=12):
    """Takagi factorization of complex symmetric matrices.

    Note:

        The singular values have to be rounded due to floating point errors.

        The result is not unique in a sense that different result could be obtained
        by different ordering of the singular values.

    References:
    - https://journals.aps.org/pra/abstract/10.1103/PhysRevA.94.062109
    """

    V, singular_values, W_adjoint = np.linalg.svd(matrix)

    W = W_adjoint.conjugate().transpose()

    singular_value_multiplicity_map = {}

    for index, value in enumerate(singular_values):
        value = np.round(value, decimals=rounding)

        if value not in singular_value_multiplicity_map:
            singular_value_multiplicity_map[value] = [index]
        else:
            singular_value_multiplicity_map[value].append(index)

    diagonal_blocks_for_Q = []

    for indices in singular_value_multiplicity_map.values():
        Z = V[:, indices].transpose() @ W[:, indices]

        diagonal_blocks_for_Q.append(scipy.linalg.sqrtm(Z))

    Q = scipy.linalg.block_diag(*diagonal_blocks_for_Q)

    return singular_values, V @ Q.conjugate()


def _rotation_to_positive_above_diagonals(block_diagonal_matrix):
    """
    The block diagonal matrix returned by the Schur decomposition in the Williamson
    decomposition needs to be rotated.

    Not doing this we'd still get a valid Williamson decompostion with valid symplectic
    and diagonal matrices, but the symplectic matrix would have complex elements and the
    diagonal matrix would have negative values.
    """

    d = len(block_diagonal_matrix) // 2
    identity = np.identity(2)
    rotation = np.rot90(identity)

    return block_diag(
        *[
            identity
            if block_diagonal_matrix[2 * index, 2 * index + 1] > 0
            else rotation
            for index in range(d)
        ]
    )


def williamson(matrix: np.ndarray) -> tuple:
    r"""
    Decomposes a positive definite matrix with Williamson decomposition, i.e. a
    positive definite :math:`M` is decomposed to

    .. math::

        M = S D S^T,

    where :math:`S \in \operatorname{Sp}(\mathbb{R}, 2d)` is a real symplectic matrix,
    and :math:`D` is a diagonal matrix containing positive values in the diagonal.

    The algorithm works as follows: without loss of generality, one can write the
    symplectic matrix in the form of

    .. math::

        S = M^{1 / 2} K D^{- 1 / 2}

    with :math:`K \in O(2d)`, since then

    .. math::

        M = S D S^T

    by construction. Now we need to find such :math:`K` that the value of :math:`S` is
    symplectic.

    .. math::

        S^T \Omega S = \Omega
            \rightleftarrow
            M^{- 1 / 2} J M^{- 1 / 2} = K D^{- 1 / 2} J D^{- 1 / 2} K^T,

    where

    .. math::

        D^{- 1 / 2} J D^{- 1 / 2}
            =
            \begin{bmatrix}
                0            & \hat{D}^{-1} \\
                \hat{D}^{-1} & 0            \\
            \end{bmatrix}

    is an antisymmetric matrix. We also know that :math:`M^{- 1 / 2} J M^{- 1 / 2}` is
    also antisymmetric. We just need to deduce the orthogonal transformation :math:`K`
    to acquire the symplectic matrix :math:`S`.

    We can use a (real) Schur decomposition to block-diagonalize
    :math:`M^{- 1 / 2} J M^{- 1 / 2}`. Note, that we also rotate the block to have the
    positive values in the above the diagonal to acquire real-valued symplectic matrices
    in the Williamson decomposition. Finally, we can acquire
    :math:`D^{- 1 / 2} J D^{- 1 / 2}` with a simple basis change.

    References:
        - https://math.stackexchange.com/a/1244793

    Args:
        matrix (numpy.ndarray): The matrix to decompose.

    Returns
        tuple: Tuple of the symplectic and diagonal matrices, in this order.
    """

    d = len(matrix) // 2

    omega = xp_symplectic_form(d)

    root_matrix = sqrtm(matrix)
    inverse_root_matrix = np.linalg.inv(root_matrix)

    block_diagonal_part, orthogonal_part = schur(
        inverse_root_matrix @ omega @ inverse_root_matrix,
        output="real",
    )

    basis_change = _rotation_to_positive_above_diagonals(
        block_diagonal_part
    ) @ from_xxpp_to_xpxp_transformation_matrix(d)
    ordered_block_diagonal = basis_change.T @ block_diagonal_part @ basis_change

    inverse_diagonal_matrix = block_diag(*(ordered_block_diagonal[:d, d:],) * 2)

    root_inverse_diagonal_matrix = np.diag(np.sqrt(np.diag(inverse_diagonal_matrix)))

    symplectic = (
        root_matrix @ orthogonal_part @ basis_change @ root_inverse_diagonal_matrix
    )

    diagonal_matrix = np.diag(1 / np.diag(inverse_diagonal_matrix))

    return symplectic, diagonal_matrix


def decompose_to_pure_and_mixed(
    matrix: np.ndarray,
    hbar: float,
) -> Tuple[np.ndarray, np.ndarray]:
    symplectic, diagonal = williamson(matrix)
    pure_covariance = hbar * symplectic @ symplectic.transpose()
    mixed_contribution = (
        symplectic
        @ (diagonal - hbar * np.identity(len(diagonal)))
        @ symplectic.transpose()
    )
    return pure_covariance, mixed_contribution



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from theboss.boson_sampling_utilities.permanent_calculators.glynn_gray_permanent_calculator import (  # noqa: E501
    GlynnGrayPermanentCalculator,
)


def _permanent(matrix, rows, columns, calculator_class):
    calculator = calculator_class(matrix, rows, columns)

    return calculator.compute_permanent()


def glynn_gray_permanent(matrix, rows, columns):
    return _permanent(
        matrix, rows, columns, calculator_class=GlynnGrayPermanentCalculator
    )



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Union

import numpy as np


def is_natural(number: Union[int, float]) -> bool:
    return bool(np.isclose(number % 1, 0.0) and round(number) >= 0)


def all_natural(array: Iterable) -> bool:
    return all(is_natural(number) for number in array)


def all_real_and_positive(vector: Iterable) -> bool:
    return all(element >= 0.0 or np.isclose(element, 0.0) for element in vector)



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Type, Optional

import abc
import copy

import numpy as np

from piquasso.api.config import Config


class State(abc.ABC):
    """The base class from which all `State` classes are derived.

    Properties:
        d (int): Instance attribute specifying the number of modes.
    """

    _config_class: Type[Config] = Config

    def __init__(self, config: Optional[Config] = None) -> None:
        self._config = config.copy() if config is not None else self._config_class()

    def _get_auxiliary_modes(self, modes: Tuple[int, ...]) -> Tuple[int, ...]:
        return tuple(np.delete(np.arange(self.d), modes))

    def copy(self) -> "State":
        """Returns an exact copy of this state.

        Returns:
            State: An exact copy of this state.
        """

        return copy.deepcopy(self)

    @property
    @abc.abstractmethod
    def d(self) -> int:
        pass

    @property
    @abc.abstractmethod
    def fock_probabilities(self) -> np.ndarray:
        """Returns the particle detection probabilities.

        Note:
            The ordering of the Fock basis is increasing with particle numbers, and in
            each particle number conserving subspace, lexicographic ordering is used.

        Returns:
            numpy.ndarray: The particle detection probabilities.
        """
        pass

    @abc.abstractmethod
    def validate(self) -> None:
        """Validates the state."""
        pass

    @abc.abstractmethod
    def get_particle_detection_probability(
        self, occupation_number: Tuple[int, ...]
    ) -> float:
        """
        Returns the particle number detection probability using the occupation number
        specified as a parameter.

        Args:
            occupation_number (tuple):
                Tuple of natural numbers representing the number of particles in each
                mode.

        Returns:
            float: The probability of detection.
        """
        pass



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import numpy as np


@functools.lru_cache()
def from_xxpp_to_xpxp_transformation_matrix(d: int) -> np.ndarray:
    r"""
    Basis changing with the basis change operator.

    This transformation will change the basis from xxpp-basis to xpxp-basis.

    .. math::

        T_{ij} = \delta_{j, 2i-1} + \delta_{j + 2d, 2i}

    Intuitively, it changes the basis as

    .. math::

        T Y = T (x_1, \dots, x_d, p_1, \dots, p_d)^T
            = (x_1, p_1, \dots, x_d, p_d)^T,

    which is very helpful in :mod:`piquasso._backends.gaussian.state`.

    Args:
        d (int): The number of modes.

    Returns:
        numpy.ndarray: The basis changing matrix.
    """

    T = np.zeros((2 * d, 2 * d), dtype=int)
    for i in range(d):
        T[2 * i, i] = 1
        T[2 * i + 1, i + d] = 1

    return T



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Mapping

import random
import numpy as np

from .state import FockState

from piquasso.api.instruction import Instruction
from piquasso.api.result import Result


def vacuum(state: FockState, instruction: Instruction, shots: int) -> Result:
    state.reset()

    return Result(state=state)


def passive_linear(state: FockState, instruction: Instruction, shots: int) -> Result:
    operator: np.ndarray = instruction._all_params["passive_block"]

    fock_operator = state._space.get_passive_fock_operator(
        operator,
        modes=instruction.modes,
        d=state._space.d,
        permanent_function=state._config.permanent_function,
    )

    state._density_matrix = (
        fock_operator @ state._density_matrix @ fock_operator.conjugate().transpose()
    )

    return Result(state=state)


def particle_number_measurement(
    state: FockState, instruction: Instruction, shots: int
) -> Result:

    reduced_state = state.reduced(instruction.modes)

    probability_map = reduced_state.fock_probabilities_map

    samples = random.choices(
        population=list(probability_map.keys()),
        weights=list(probability_map.values()),
        k=shots,
    )

    # NOTE: We choose the last sample for multiple shots.
    sample = samples[-1]

    normalization = _get_normalization(probability_map, sample)

    _project_to_subspace(
        state=state,
        subspace_basis=sample,
        modes=instruction.modes,
        normalization=normalization,
    )

    return Result(state=state, samples=samples)


def _get_normalization(
    probability_map: Mapping[Tuple[int, ...], float], sample: Tuple[int, ...]
) -> float:
    return 1 / probability_map[sample]


def _project_to_subspace(
    state: FockState,
    *,
    subspace_basis: Tuple[int, ...],
    modes: Tuple[int, ...],
    normalization: float
) -> None:
    projected_density_matrix = _get_projected_density_matrix(
        state=state,
        subspace_basis=subspace_basis,
        modes=modes,
    )

    state._density_matrix = projected_density_matrix * normalization


def _get_projected_density_matrix(
    state: FockState, *, subspace_basis: Tuple[int, ...], modes: Tuple[int, ...]
) -> np.ndarray:
    new_density_matrix = state._get_empty()

    index = state._space.get_projection_operator_indices(
        subspace_basis=subspace_basis,
        modes=modes,
    )

    new_density_matrix[index] = state._density_matrix[index]

    return new_density_matrix


def create(state: FockState, instruction: Instruction, shots: int) -> Result:
    operator = state._space.get_creation_operator(instruction.modes)

    state._density_matrix = operator @ state._density_matrix @ operator.transpose()

    state.normalize()

    return Result(state=state)


def annihilate(state: FockState, instruction: Instruction, shots: int) -> Result:
    operator = state._space.get_annihilation_operator(instruction.modes)

    state._density_matrix = operator @ state._density_matrix @ operator.transpose()

    state.normalize()

    return Result(state=state)


def kerr(state: FockState, instruction: Instruction, shots: int) -> Result:
    mode = instruction.modes[0]
    xi = instruction._all_params["xi"]
    for x in xi:
        for index, (basis, dual_basis) in state._space.operator_basis:
            number = basis[mode]
            dual_number = dual_basis[mode]

            coefficient = np.exp(
                1j
                * x
                * (number * (2 * number + 1) - dual_number * (2 * dual_number + 1))
            )

            state._density_matrix[index] *= coefficient

    return Result(state=state)


def cubic_phase(state: FockState, instruction: Instruction, shots: int) -> Result:
    gamma = instruction._all_params["gamma"]
    hbar = state._config.hbar

    for index, mode in enumerate(instruction.modes):
        operator = state._space.get_single_mode_cubic_phase_operator(
            gamma=gamma[index], hbar=hbar
        )
        embedded_operator = state._space.embed_matrix(
            operator,
            modes=(mode,),
            auxiliary_modes=state._get_auxiliary_modes(instruction.modes),
        )
        state._density_matrix = (
            embedded_operator
            @ state._density_matrix
            @ embedded_operator.conjugate().transpose()
        )

        state.normalize()

    return Result(state=state)


def cross_kerr(state: FockState, instruction: Instruction, shots: int) -> Result:
    modes = instruction.modes
    xi = instruction._all_params["xi"]

    for index, (basis, dual_basis) in state._space.operator_basis:
        coefficient = np.exp(
            1j
            * xi
            * (
                basis[modes[0]] * basis[modes[1]]
                - dual_basis[modes[0]] * dual_basis[modes[1]]
            )
        )

        state._density_matrix[index] *= coefficient

    return Result(state=state)


def displacement(state: FockState, instruction: Instruction, shots: int) -> Result:
    amplitudes = np.abs(instruction._all_params["displacement_vector"])
    angles = np.angle(instruction._all_params["displacement_vector"])

    for index, mode in enumerate(instruction.modes):
        operator = state._space.get_single_mode_displacement_operator(
            r=amplitudes[index],
            phi=angles[index],
        )

        embedded_operator = state._space.embed_matrix(
            operator,
            modes=(mode,),
            auxiliary_modes=state._get_auxiliary_modes(instruction.modes),
        )

        state._density_matrix = (
            embedded_operator
            @ state._density_matrix
            @ embedded_operator.conjugate().transpose()
        )

        state.normalize()

    return Result(state=state)


def squeezing(state: FockState, instruction: Instruction, shots: int) -> Result:
    amplitudes = np.arccosh(np.diag(instruction._all_params["passive_block"]))
    angles = np.angle(-np.diag(instruction._all_params["active_block"]))

    for index, mode in enumerate(instruction.modes):
        operator = state._space.get_single_mode_squeezing_operator(
            r=amplitudes[index],
            phi=angles[index],
        )

        embedded_operator = state._space.embed_matrix(
            operator,
            modes=(mode,),
            auxiliary_modes=state._get_auxiliary_modes(instruction.modes),
        )

        state._density_matrix = (
            embedded_operator
            @ state._density_matrix
            @ embedded_operator.conjugate().transpose()
        )

        state.normalize()

    return Result(state=state)


def linear(state: FockState, instruction: Instruction, shots: int) -> Result:
    operator = state._space.get_linear_fock_operator(
        modes=instruction.modes,
        passive_block=instruction._all_params["passive_block"],
        active_block=instruction._all_params["active_block"],
        permanent_function=state._config.permanent_function,
    )

    state._density_matrix = (
        operator @ state._density_matrix @ operator.conjugate().transpose()
    )

    state.normalize()

    return Result(state=state)


def density_matrix_instruction(
    state: FockState, instruction: Instruction, shots: int
) -> Result:
    _add_occupation_number_basis(state, **instruction.params)

    return Result(state=state)


def _add_occupation_number_basis(
    state: FockState,
    *,
    ket: Tuple[int, ...],
    bra: Tuple[int, ...],
    coefficient: complex
) -> None:
    index = state._space.index(ket)
    dual_index = state._space.index(bra)

    state._density_matrix[index, dual_index] = coefficient



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Tuple

import numpy as np

from .validations import all_real_and_positive


def is_unitary(matrix: np.ndarray) -> bool:
    return np.allclose(
        matrix @ matrix.conjugate().transpose(), np.identity(matrix.shape[0])
    )


def is_real(matrix: np.ndarray) -> bool:
    return bool(np.all(np.isreal(matrix)))


def is_2n_by_2n(matrix: np.ndarray) -> bool:
    dim = len(matrix)

    return matrix.shape == (dim, dim) and dim % 2 == 0


def is_real_2n_by_2n(matrix: np.ndarray) -> bool:
    return is_real(matrix) and is_2n_by_2n(matrix)


def is_symmetric(matrix: np.ndarray) -> bool:
    return np.allclose(matrix, matrix.transpose())


def is_selfadjoint(matrix: np.ndarray) -> bool:
    return np.allclose(matrix, matrix.conjugate().transpose())


def is_positive_semidefinite(matrix: np.ndarray) -> bool:
    eigenvalues = np.linalg.eigvals(matrix)

    return all_real_and_positive(eigenvalues)


def is_square(matrix: np.ndarray) -> bool:
    shape = matrix.shape
    return len(shape) == 2 and shape[0] == shape[1]


def is_invertible(matrix: np.ndarray) -> bool:
    return is_square(matrix) and np.linalg.matrix_rank(matrix) == matrix.shape[0]


def is_diagonal(matrix: np.ndarray) -> bool:
    return np.allclose(matrix, np.diag(np.diag(matrix)))


def reduce_(array: np.ndarray, reduce_on: Iterable[int]) -> np.ndarray:
    proper_index = []

    for index, multiplier in enumerate(reduce_on):
        proper_index.extend([index] * multiplier)

    if array.ndim == 1:
        return array[proper_index]

    return array[np.ix_(proper_index, proper_index)]


def block_reduce(array: np.ndarray, reduce_on: Tuple[int, ...]) -> np.ndarray:
    return reduce_(array, reduce_on=(reduce_on * 2))



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import numpy as np


def get_operator_index(modes: Tuple[int, ...]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Note:
        For indexing of numpy arrays, see
        https://numpy.org/doc/stable/reference/arrays.indexing.html#advanced-indexing
    """

    transformed_columns = np.array([modes] * len(modes))
    transformed_rows = transformed_columns.transpose()

    return transformed_rows, transformed_columns


def get_auxiliary_operator_index(
    modes: Tuple[int, ...], auxiliary_modes: Tuple[int, ...]
) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
    auxiliary_rows = tuple(np.array([modes] * len(auxiliary_modes)).transpose())

    return auxiliary_rows, auxiliary_modes



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, List

import numpy as np
from scipy.linalg import sqrtm

from piquasso.api.config import Config
from piquasso.api.errors import InvalidState, InvalidParameter, PiquassoException
from piquasso.api.state import State

from piquasso._math.functions import gaussian_wigner_function
from piquasso._math.linalg import (
    is_symmetric,
    is_positive_semidefinite,
)
from piquasso._math.symplectic import symplectic_form, xp_symplectic_form
from piquasso._math.combinatorics import get_occupation_numbers
from piquasso._math.transformations import from_xxpp_to_xpxp_transformation_matrix

from .probabilities import DensityMatrixCalculation


class GaussianState(State):
    r"""Class to represent a Gaussian state."""

    def __init__(self, d: int, config: Config = None) -> None:
        """
        Args:
            d (int): The number of modes.
        """
        super().__init__(config=config)
        self._d = d
        self.reset()

    def __len__(self) -> int:
        return self._d

    @property
    def d(self) -> int:
        return len(self)

    def reset(self) -> None:
        r"""Resets the state to a vacuum."""

        vector_shape = (self.d,)
        matrix_shape = vector_shape * 2

        self._m = np.zeros(vector_shape, dtype=complex)
        self._G = np.zeros(matrix_shape, dtype=complex)
        self._C = np.zeros(matrix_shape, dtype=complex)

    @classmethod
    def _from_representation(
        cls,
        *,
        m: np.ndarray,
        G: np.ndarray,
        C: np.ndarray,
        config: Config,
    ) -> "GaussianState":
        obj = cls(d=len(m), config=config)

        obj._m = m
        obj._G = G
        obj._C = C

        return obj

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GaussianState):
            return False

        return (
            np.allclose(self._C, other._C)
            and np.allclose(self._G, other._G)
            and np.allclose(self._m, other._m)
        )

    def validate(self) -> None:
        r"""
        Validates the state.

        Raises:
            InvalidState:
                Raised if the underlying Gaussian state is invalid, which could mean
                - ill-shaped mean and covariance;
                - non-symmetric covariance matrix;
                - the covariance matrix doesn't fulfill the Robertson-Schrödinger
                  uncertainty relations.
        """

        self._validate_mean(self.xpxp_mean_vector, self.d)
        self._validate_cov(self.xpxp_covariance_matrix, self.d)

    @staticmethod
    def _validate_mean(mean: np.ndarray, d: int) -> None:
        expected_shape = (2 * d,)

        if not mean.shape == expected_shape:
            raise InvalidState(
                f"Invalid 'mean' vector shape; "
                f"expected={expected_shape}, actual={mean.shape}."
            )

    def _validate_cov(self, cov: np.ndarray, d: int) -> None:
        expected_shape = (2 * d,) * 2

        if not cov.shape == expected_shape:
            raise InvalidState(
                f"Invalid 'cov' matrix shape; "
                f"expected={expected_shape}, actual={cov.shape}."
            )

        if not is_symmetric(cov):
            raise InvalidState("The covariance matrix is not symmetric.")

        if not is_positive_semidefinite(
            cov / self._config.hbar + 1j * symplectic_form(d)
        ):
            raise InvalidState(
                "The covariance matrix is invalid, since it doesn't fulfill the "
                "Robertson-Schrödinger uncertainty relation."
            )

    @property
    def xxpp_mean_vector(self) -> np.ndarray:
        r"""The state's mean in the xxpp-ordered basis.

        The expectation value of the quadrature operators in xxpp basis, i.e.
        :math:`\operatorname{Tr} \left ( \rho Y \right )`, where
        :math:`Y = (x_1, \dots, x_d, p_1, \dots, p_d)^T`.

        Returns:
            numpy.ndarray: A :math:`2d`-vector where :math:`d` is the number of modes.
        """

        dimensionless_xxpp_mean_vector = np.concatenate(
            [self._m.real, self._m.imag]
        ) * np.sqrt(2)

        return dimensionless_xxpp_mean_vector * np.sqrt(self._config.hbar)

    @xxpp_mean_vector.setter
    def xxpp_mean_vector(self, value):
        T = from_xxpp_to_xpxp_transformation_matrix(self.d)
        self.xpxp_mean_vector = T @ value

    @property
    def xxpp_covariance_matrix(self) -> np.ndarray:
        r"""The xxpp-ordered coveriance matrix of the state.

        The xxpp-ordered covariance matrix :math:`\sigma_{xp}` is defined by

        .. math::
            \sigma_{xp, ij} := \langle Y_i Y_j + Y_j Y_i \rangle_\rho
                - 2 \langle Y_i \rangle_\rho \langle Y_j \rangle_\rho,

        where

        .. math::
            Y = (x_1, \dots, x_d, p_1, \dots, p_d)^T,

        and :math:`\rho` is the density operator of the currently represented state.

        Returns:
            numpy.ndarray:
                The :math:`2d \times 2d` xp-ordered covariance matrix.
        """

        C = self._C
        G = self._G

        dimensionless_xxpp_covariance_matrix = 2 * np.block(
            [
                [(G + C).real, (G + C).imag],
                [(G - C).imag, (-G + C).real],
            ],
        ) + np.identity(2 * self.d)

        return dimensionless_xxpp_covariance_matrix * self._config.hbar

    @xxpp_covariance_matrix.setter
    def xxpp_covariance_matrix(self, value):
        T = from_xxpp_to_xpxp_transformation_matrix(self.d)
        self.xpxp_covariance_matrix = T @ value @ T.transpose()

    @property
    def xxpp_correlation_matrix(self) -> np.ndarray:
        r"""The state's correlation matrix in the xxpp basis.

        Let :math:`M_{(xxpp)}` be the correlation matrix in the xxpp basis.
        Then

        .. math::
            M_{ij (xxpp)} = \langle Y_i Y_j + Y_j Y_i \rangle_\rho,

        where :math:`M_{ij (xxpp)}` denotes the matrix element in the
        :math:`i`-th row and :math:`j`-th column,

        .. math::
            Y = (x_1, \dots, x_d, p_1, \dots, p_d)^T,

        and :math:`\rho` is the density operator of the currently represented state.

        Returns:
            numpy.ndarray: The :math:`2d \times 2d` correlation matrix in the
            xxpp-basis.
        """
        xxpp_mean_vector = self.xxpp_mean_vector
        return self.xxpp_covariance_matrix + 2 * np.outer(
            xxpp_mean_vector, xxpp_mean_vector
        )

    @property
    def xxpp_representation(self) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        The state's mean and correlation matrix ordered in the xxpp basis.

        Returns:
            tuple: :meth:`xxpp_mean_vector`, :meth:`xxpp_correlation_matrix`.
        """

        return self.xxpp_mean_vector, self.xxpp_correlation_matrix

    @property
    def xpxp_mean_vector(self) -> np.ndarray:
        r"""Returns the xpxp-ordered mean of the state.

        Returns:
            numpy.ndarray: A :math:`2d`-vector. The expectation value of the quadrature
            operators in `xxpp`-ordering, i.e. :math:`\operatorname{Tr} \rho R`, where
            :math:`R = (x_1, p_1, \dots, x_d, p_d)^T`.
        """
        T = from_xxpp_to_xpxp_transformation_matrix(self.d)
        return T @ self.xxpp_mean_vector

    @xpxp_mean_vector.setter
    def xpxp_mean_vector(self, value: np.ndarray) -> None:
        self._validate_mean(value, self.d)

        m = (value[::2] + 1j * value[1::2]) / np.sqrt(2 * self._config.hbar)

        self._m = m

    @property
    def xpxp_covariance_matrix(self) -> np.ndarray:
        r"""The `xpxp`-ordered coveriance matrix of the state.

        The `xpxp`-ordered covariance matrix :math:`\sigma` is defined by

        .. math::
            \sigma_{ij} := \langle R_i R_j + R_j R_i \rangle_\rho
                - 2 \langle R_i \rangle_\rho \langle R_j \rangle_\rho,

        where

        .. math::
            R = (x_1, p_1, \dots, x_d, p_d)^T,

        and :math:`\rho` is the density operator of the currently represented state.

        Returns:
            numpy.ndarray:
                The :math:`2d \times 2d` `xpxp`-ordered covariance matrix in
                xpxp-ordered basis.
        """

        T = from_xxpp_to_xpxp_transformation_matrix(self.d)
        return T @ self.xxpp_covariance_matrix @ T.transpose()

    @xpxp_covariance_matrix.setter
    def xpxp_covariance_matrix(self, new_cov: np.ndarray) -> None:
        d = self.d

        self._validate_cov(new_cov, d)

        T = from_xxpp_to_xpxp_transformation_matrix(d)

        dimensionless_cov = new_cov / self._config.hbar
        dimensionless_xp_cov = T.transpose() @ dimensionless_cov @ T

        blocks = (dimensionless_xp_cov - np.identity(2 * d)) / 4

        C_real = blocks[:d, :d] + blocks[d:, d:]
        G_real = blocks[:d, :d] - blocks[d:, d:]

        C_imag = blocks[:d, d:] - blocks[d:, :d]
        G_imag = blocks[:d, d:] + blocks[d:, :d]

        C = C_real + 1j * C_imag
        G = G_real + 1j * G_imag

        self._G = G
        self._C = C

    @property
    def xpxp_correlation_matrix(self) -> np.ndarray:
        r"""The `xpxp`-ordered correlation matrix of the state.

        Let :math:`M` be the correlation matrix in the `xpxp` basis.
        Then

        .. math::
            M_{ij} = \langle R_i R_j + R_j R_i \rangle_\rho,

        where :math:`M_{ij}` denotes the matrix element in the
        :math:`i`-th row and :math:`j`-th column,

        .. math::
            R = (x_1, p_1, \dots, x_d, p_d)^T,

        and :math:`\rho` is the density operator of the currently represented state.

        Returns:
            numpy.ndarray: The :math:`2d \times 2d` `xpxp`-ordered correlation matrix.
        """

        T = from_xxpp_to_xpxp_transformation_matrix(self.d)
        return T @ self.xxpp_correlation_matrix @ T.transpose()

    @property
    def xpxp_representation(self) -> Tuple[np.ndarray, np.ndarray]:
        r"""The state's mean and correlation matrix ordered by the `xpxp` basis.

        Returns:
            tuple: :meth:`mean`, :meth:`corr`.
        """

        return self.xpxp_mean_vector, self.xpxp_correlation_matrix

    @property
    def complex_displacement(self) -> np.ndarray:
        r"""The complex displacement of the state.

        The complex displacement is defined by

        .. math::
            \mu_{c} := \begin{bmatrix}
                \langle a_1 \rangle_{\rho}, \dots, \langle a_d \rangle_{\rho},
                \langle a^\dagger_1 \rangle_{\rho},
                \dots, \langle a^\dagger_d \rangle_{\rho}
            \end{bmatrix}.

        Equivalently, one can write

        .. math::
            \mu_{c} := W \mu_{xxpp},

        where :math:`\mu_{xxpp}` is the xxpp-ordered mean vector
        :attr:`xxpp_mean_vector` and

        .. math::
            W = \frac{1}{\sqrt{2}} \begin{bmatrix}
                I_{d} & i I_{d} \\
                I_{d} & -i I_{d}
            \end{bmatrix}.

        Returns:
            numpy.ndarray: The complex displacement.
        """

        return np.concatenate([self._m, self._m.conj()])

    @property
    def complex_covariance(self) -> np.ndarray:
        r"""The complex covariance of the state.

        The complex covariance is defined by

        .. math::
            \sigma_{c, ij} = \langle
                \xi_i \xi^\dagger_j + \xi_j \xi^\dagger_i
            \rangle_{\rho}
            - 2 \langle \xi_i \rangle_{\rho} \langle \xi^\dagger_j \rangle_{\rho},

        where

        .. math::
            \xi = \begin{bmatrix}
                a_1, \dots a_d, a^\dagger_1, \dots, a^\dagger_d
            \end{bmatrix}.

        Equivalently, one can write

        .. math::
            \sigma_{c} = \frac{1}{\hbar} W \sigma_{xxpp} W^{\dagger},

        where :math:`\sigma_{xxpp}` is the xxpp-ordered covariance
        matrix :attr:`xxpp_cov` and

        .. math::
            W = \frac{1}{\sqrt{2}} \begin{bmatrix}
                I_{d} & i I_{d} \\
                I_{d} & -i I_{d}
            \end{bmatrix}.

        Returns:
            numpy.ndarray: The complex covariance.
        """

        return 2 * np.block(
            [[self._C.conj(), self._G], [self._G.conj(), self._C]]
        ) + np.identity(2 * self.d)

    @property
    def Q_matrix(self):
        return (self.complex_covariance + np.identity(2 * len(self))) / 2

    def rotated(self, phi: float) -> "GaussianState":
        r"""Returns a copy of the current Gaussian state, rotated by an angle `phi`.

        Let :math:`\phi \in [ 0, 2 \pi )`. Let us define the following:

        .. math::
            x_{i, \phi} = \cos\phi~x_i + \sin\phi~p_i,

        which is a generalized quadrature operator. One could rotate the state by a
        simple complex phase, which can be shown by using the transformation rules
        between the ladder operators and quadrature operators, i.e.

        .. math::
            x_i &= \sqrt{\frac{\hbar}{2}} (a_i + a_i^\dagger) \\
            p_i &= -i \sqrt{\frac{\hbar}{2}} (a_i - a_i^\dagger),

        which we could rewrite :math:`x_{i, \phi}` to

        .. math::
            x_{i, \phi} = \sqrt{\frac{\hbar}{2}} \left(
                a_i \exp(-i \phi) + a_i^\dagger \exp(i \phi)
            \right),

        meaning that e.g. the annihilation operators :math:`a_i` are transformed just
        by multiplying it with a phase factor :math:`\exp(-i \phi)` i.e.

        .. math::
            (\langle a_i \rangle_{\rho} =: )~m_i &\mapsto \exp(-i \phi) m_i \\
            (\langle a^\dagger_i a_j \rangle_{\rho} =: )~C_{ij} &\mapsto C_{ij} \\
            (\langle a_i a_j \rangle_{\rho} =: )~G_{ij} &\mapsto \exp(-i 2 \phi) G_{ij}.

        Args:
            phi (float): The angle to rotate the state with.

        Returns:
            GaussianState: The rotated `GaussianState` instance.
        """
        phase = np.exp(-1j * phi)

        return self.__class__._from_representation(
            C=self._C,
            G=(self._G * phase**2),
            m=(self._m * phase),
            config=self._config,
        )

    def reduced(self, modes: Tuple[int, ...]) -> "GaussianState":
        """Returns a copy of the current state, reduced to the given `modes`.

        This method essentially preserves the modes specified from the representation
        of the Gaussian state, but cuts out the other modes.

        Args:
            modes (tuple[int]): The modes to reduce the state to.

        Returns:
            GaussianState: The reduced `GaussianState` instance.
        """
        return self.__class__._from_representation(
            C=self._C[np.ix_(modes, modes)],
            G=self._G[np.ix_(modes, modes)],
            m=self._m[np.ix_(modes)],
            config=self._config,
        )

    def xpxp_reduced_rotated_mean_and_covariance(
        self, modes: Tuple[int, ...], phi: float
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""The mean and covariance on a rotated and reduced state.

        Let the index set :math:`\vec{i}` correspond to `modes`, and the angle
        :math:`\phi` correspond to `phi`. The current :class:`GaussianState` instance
        is reduced to `modes` and rotated by `phi` in a new instance, and let that
        state be denoted by :math:`\rho_{\vec{i}, \phi}`.

        Then the `xpxp`-ordered mean and covariance can be calculated by

        .. math::
            \mu_{\vec{i}, \phi}
                &:= \langle R_{\vec{i}} \rangle_{\rho_{\vec{i}, \phi}}, \\
            \sigma_{\vec{i}, \phi}
                &:=  \langle
                    R_{\vec{i}} R_{\vec{i}}^T
                \rangle_{\rho_{\vec{i}, \phi}}
                - \mu_{\vec{i}, \phi} \mu_{\vec{i}, \phi}^T,

        where

        .. math::
            R = (x_1, p_1, \dots, x_d, p_d)^T,

        and :math:`R_{\vec{i}}` is just the same vector, reduced to a subsystem
        specified by :math:`\vec{i}`.

        Args:
            modes (tuple[int]): The modes to reduce the state to.
            phi (float): The angle to rotate the state with.

        Returns:
            (numpy.ndarray, numpy.ndarray):
                `xpxp`-ordered mean and covariance of the reduced and rotated
                version of the current :class:`GaussianState`.
        """
        transformed_state = self.reduced(modes).rotated(phi)

        return (
            transformed_state.xpxp_mean_vector,
            transformed_state.xpxp_covariance_matrix,
        )

    def mean_photon_number(self, modes: Tuple[int, ...]) -> float:
        r"""This method returns the mean photon number of the given modes.
        The mean photon number :math:`\bar{n} = \langle \hat{n}
        \rangle` can be calculated in terms of the ladder operators by the following
        expression

        .. math::
            \sum_{i=1}^{d} \langle a_{i}^\dagger a_{i} \rangle =
            \operatorname{Tr}(\rho \hat{n}),

        where :math:`a`, :math:`a ^\dagger` are the annihilation and the creation
        operators respectively, :math:`\rho` is the density operator of the
        currently represented state and :math:`d` is the number of modes. For a general
        displaced squeezed gaussian state, the mean photon number is

        .. math::
            \langle \hat{n} \rangle = \operatorname{Tr}(\langle a^\dagger a \rangle) +
            \mu_{c}^ \dagger \cdot \mu_{c},

        where :math:`\mu_{c}` is the :attr:`complex_displacement`.

        .. note::
            This method can also be used to return the summation of the mean photon
            number for multiple modes if the mode parameter contains more than one
            integer e.g :math:`(0,1,...)`.

        Args:
            modes (tuple[int]): The correspoding modes at which the mean photon number
                is calculated.
        Returns:
            float: The expectation value of the photon number.
        """

        state = self.reduced(modes)
        return (np.trace(state._C) + state._m.conjugate() @ state._m).real

    def variance_photon_number(self, modes: Tuple[int, ...]) -> float:
        r"""
        This method calculates the variance of the photon number operator as follows:

        .. math::
            \operatorname{\textit{Var(n)}} = \frac{1}{2} \operatorname{Tr}(
                \sigma_{ij}/2\hbar)^2 +  \det (\sigma_{ij}/2\hbar) +
                \frac{1}{2} \left \langle Q \right \rangle (\sigma_{ij}/2\hbar)
                \left \langle Q \right \rangle - 0.25

        where :math:`\sigma_{ij}` is the reduced :attr:`xpxp_covariance_matrix` since
        :math:`i = j` i.e. the covariance matrix of the reduced Gaussian state and
        :math:`\left \langle Q \right \rangle` is the :attr:`xpxp_mean_vector`.

        For more details please see
        `this article <https://arxiv.org/pdf/1612.01199.pdf>`_.

        Args:
            modes (Tuple[int, ...]): The correspoding modes at which the variance of
                the photon number is calculated.

        Returns:
            float: Variance of the photon number operator
        """
        reduced_state = self.reduced(modes=modes)
        mean, covariance = (
            reduced_state.xpxp_mean_vector,
            reduced_state.xpxp_covariance_matrix / (2 * self._config.hbar),
        )

        return (
            0.5 * np.trace(covariance) ** 2
            - np.linalg.det(covariance)
            - 0.25
            + (0.5 * mean @ covariance @ mean)
        )

    def fidelity(self, state: "GaussianState") -> float:
        r"""Calculates the state fidelity between two quantum states.

        The state fidelity :math:`F` between two density matrices
        :math:`\rho_1, \rho_2` is given by:

        .. math::
            \operatorname{F}(\rho_1, \rho_2) = \operatorname{Tr}(\sqrt{\sqrt{\rho_1}
                \rho_2\sqrt{\rho_1}})^2

        A gaussian state can be represented by its Covariance matrix and the vector of
        Means. Hence, the above equation can be rewritten as:

        .. math::
            \operatorname{F} = \operatorname{F_0}(V_1, V_2) \exp(
                -\frac{1}{2} \delta_u^T(V_1 + V_2)^{-1}\delta_u^T)

        where :math:`V` is the :attr:`xpxp_covariance_matrix` of the gaussian state,
        :math:`\delta_u` is the difference between mean vectors of the two gaussian
        states represented by :attr:`xpxp_mean_vector`, and :math:`F_0` is given by:

        .. math::
            \operatorname{F_0} = \sqrt{\det{[2(\sqrt{I +
                \frac{(V_{aux}\Omega)^-2}{4}} + I)V_{aux}]} \det{[(V_1 + V_2)^{-1}]}}

        where :math:`V_{aux}` is given by

        .. math::
            \Omega^T (V_1 + V_2)^{-1} (\frac{\Omega}{4} V_2 \Omega V_1)

        and :math:`\Omega` is a symplectic matrix of shape :math:`2*d \times 2*d`.
        For more details please check:
        https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.115.260501.

        Args:
            state: A gaussian state
                :class:`~piquasso._backends.gaussian.state.GaussianState` that can be
                used to calculate the fidelity aganist it.

        Returns:
            float: The calculated fidelity.
        """
        mean_1, cov_1 = (
            self.xpxp_mean_vector,
            self.xpxp_covariance_matrix / (2 * self._config.hbar),
        )
        mean_2, cov_2 = (
            state.xpxp_mean_vector,
            state.xpxp_covariance_matrix / (2 * state._config.hbar),
        )

        W = xp_symplectic_form(self.d)
        ident = np.identity(self.d * 2, dtype=complex)

        sum_cov_inv = np.linalg.inv(cov_1 + cov_2)
        V_aux = W.T @ sum_cov_inv @ (0.25 * W + cov_2 @ W @ cov_1)
        delta_mu = (mean_1 - mean_2) / np.sqrt(self._config.hbar)

        f1 = np.exp(-0.5 * delta_mu @ sum_cov_inv @ delta_mu)
        f_total = (
            2
            * (sqrtm(ident + 0.25 * np.linalg.inv(V_aux @ W @ V_aux @ W)) + ident)
            @ V_aux
        )
        f_total = np.sqrt(np.linalg.det(f_total) * np.linalg.det(sum_cov_inv))

        return float((f_total * f1).real)

    def quadratic_polynomial_expectation(
        self, A: np.ndarray, b: np.ndarray, c: float = 0.0, phi: float = 0.0
    ) -> float:
        r"""The expectation value of the specified quadratic polynomial.

        A quadratic polynomial can be written as

        .. math::
            f(R) = R^T A R + R \cdot b + c,

        where :math:`R = (x_1, p_1, \dots, x_d, p_d)^T` is the vector of the quadrature
        operators where :math:`d` is the number of modes,
        :math:`A \in \mathbb{R}^{2d \times 2d}` is a symmetric matrix,
        :math:`b \in \mathbb{R}^{2d}`, and :math:`c\in\mathbb{R}`.

        This method returns the expectation value :math:`E[f(R)]` using the following
        equation:

        .. math::
            \operatorname{E}[f(R)]
                = \operatorname{Tr}[ A\sigma ] + \mu^T A \mu + \mu^T b + c,

        where :math:`\sigma` is the covariance matrix, :math:`\mu = E[R]` is the mean
        of the quadrature operators and

        .. math::
            \operatorname{E}[\cdot] = \operatorname{Tr}[\cdot \rho],

        where :math:`\rho` is the density matrix of the state.

        Args:
            A (numpy.ndarray):
                A :math:`2d \times 2d` real symmetric matrix corresponding to the
                quadratic coefficients, where :math:`d` is the number of modes.
            b (numpy.ndarray):
                A one-dimensional :math:`2d`-length real-valued vector that corresponds
                to the first order terms of the quadratic polynomial.
            c (float): The constant term in the quadratic polynomial. Defaults to `0`.
            phi (float): Rotation angle, by which the state is rotated. Defaults to `0`.
        Returns:
            float: The expectation value of the quadratic polynomial.
        """

        if not is_symmetric(A):
            raise InvalidParameter("The specified matrix is not symmetric.")

        state = self.rotated(phi)
        mean = state.xpxp_mean_vector
        cov = state.xpxp_covariance_matrix
        first_moment = np.trace(A @ cov) / 2 + mean @ A @ mean + mean @ b + c
        # TODO: calculate the variance.
        return first_moment

    @property
    def density_matrix(self) -> np.ndarray:
        calculation = DensityMatrixCalculation(
            complex_displacement=self.complex_displacement,
            complex_covariance=self.complex_covariance,
            loop_hafnian_function=self._config.loop_hafnian_function,
        )

        return calculation.get_density_matrix(
            get_occupation_numbers(d=self.d, cutoff=self._config.cutoff)
        )

    def wigner_function(
        self,
        positions: List[List[float]],
        momentums: List[List[float]],
        modes: Tuple[int, ...] = None,
    ) -> np.ndarray:
        r"""
        This method calculates the Wigner function values at the specified position and
        momentum vectors, according to the following equation:

        .. math::
            W(r) = \frac{1}{\pi^d \sqrt{\mathrm{det} \sigma}}
                \exp \big (
                    - (r - \mu)^T
                    \sigma^{-1}
                    (r - \mu)
                \big ).

        Args:
            positions (list[list[float]]): List of position vectors.
            momentums (list[list[float]]): List of momentum vectors.
            modes (tuple[int], optional):
                Modes where Wigner function should be calculcated.

        Returns:
            numpy.ndarray:
                The Wigner function values in the shape of a grid specified by the
                input.
        """

        if modes:
            reduced_state = self.reduced(modes)
            return gaussian_wigner_function(
                positions,
                momentums,
                d=reduced_state.d,
                mean=reduced_state.xpxp_mean_vector,
                cov=reduced_state.xpxp_covariance_matrix,
            )

        return gaussian_wigner_function(
            positions,
            momentums,
            d=self.d,
            mean=self.xpxp_mean_vector,
            cov=self.xpxp_covariance_matrix,
        )

    def get_particle_detection_probability(
        self, occupation_number: Tuple[int, ...]
    ) -> float:
        if len(occupation_number) != self.d:
            raise PiquassoException(
                f"The specified occupation number should have length '{self.d}': "
                f"occupation_number='{occupation_number}'."
            )

        calculation = DensityMatrixCalculation(
            self.complex_displacement,
            self.complex_covariance,
            loop_hafnian_function=self._config.loop_hafnian_function,
        )

        return calculation.get_density_matrix_element(
            bra=occupation_number,
            ket=occupation_number,
        )

    @property
    def fock_probabilities(self) -> np.ndarray:
        calculation = DensityMatrixCalculation(
            complex_displacement=self.complex_displacement,
            complex_covariance=self.complex_covariance,
            loop_hafnian_function=self._config.loop_hafnian_function,
        )

        return calculation.get_particle_number_detection_probabilities(
            get_occupation_numbers(d=self.d, cutoff=self._config.cutoff)
        )



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Piquasso module.

One can access all the instructions and states from here as attributes.
"""

from piquasso.api.mode import Q
from piquasso.api.config import Config
from piquasso.api.instruction import (
    Instruction,
    Preparation,
    Gate,
    Measurement,
)
from piquasso.api.program import Program
from piquasso.api.state import State
from piquasso.api.computer import Computer
from piquasso.api.simulator import Simulator
from piquasso.api.utils import as_code

from piquasso._backends.sampling import SamplingState, SamplingSimulator

from piquasso._backends.gaussian import GaussianState, GaussianSimulator
from piquasso._backends.fock import (
    FockState,
    PureFockState,
    FockSimulator,
    PureFockSimulator,
)

from .instructions.preparations import (
    Vacuum,
    Mean,
    Covariance,
    Thermal,
    StateVector,
    DensityMatrix,
    Create,
    Annihilate,
)

from .instructions.gates import (
    GaussianTransform,
    Phaseshifter,
    Beamsplitter,
    MachZehnder,
    Fourier,
    Displacement,
    PositionDisplacement,
    MomentumDisplacement,
    Squeezing,
    QuadraticPhase,
    Squeezing2,
    Kerr,
    CrossKerr,
    ControlledX,
    ControlledZ,
    Interferometer,
    Graph,
    CubicPhase,
)

from .instructions.measurements import (
    ParticleNumberMeasurement,
    ThresholdMeasurement,
    HomodyneMeasurement,
    HeterodyneMeasurement,
    GeneraldyneMeasurement,
    Sampling,
)

from .instructions.channels import (
    DeterministicGaussianChannel,
    Attenuator,
    Loss,
)


__all__ = [
    # API
    "Program",
    "Q",
    "Config",
    "Instruction",
    "Preparation",
    "Gate",
    "Measurement",
    "State",
    "Computer",
    "Simulator",
    "as_code",
    # Simulators
    "GaussianSimulator",
    "SamplingSimulator",
    "FockSimulator",
    "PureFockSimulator",
    # States
    "GaussianState",
    "SamplingState",
    "FockState",
    "PureFockState",
    # Preparations
    "Vacuum",
    "Mean",
    "Covariance",
    "Thermal",
    "StateVector",
    "DensityMatrix",
    "Create",
    "Annihilate",
    # Gates
    "GaussianTransform",
    "Phaseshifter",
    "Beamsplitter",
    "MachZehnder",
    "Fourier",
    "Displacement",
    "PositionDisplacement",
    "MomentumDisplacement",
    "Squeezing",
    "QuadraticPhase",
    "Squeezing2",
    "Kerr",
    "CrossKerr",
    "CubicPhase",
    "ControlledX",
    "ControlledZ",
    "Interferometer",
    "Sampling",
    "Graph",
    # Measurements
    "ParticleNumberMeasurement",
    "ThresholdMeasurement",
    "HomodyneMeasurement",
    "HeterodyneMeasurement",
    "GeneraldyneMeasurement",
    # Channels
    "DeterministicGaussianChannel",
    "Attenuator",
    "Loss",
]

__version__ = "0.8.2"



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from typing import Collection, Tuple, Union, Any

from piquasso.core import _context

from .errors import InvalidModes, PiquassoException

if typing.TYPE_CHECKING:
    from piquasso.api.instruction import Instruction
    from piquasso.api.program import Program


class Q:
    """
    The implementation of qumodes, which is used to track on which qumodes are
    the operators placed in the circuit.

    Example usage::

        import numpy as np
        import piquasso as pq

        with pq.Program() as program:
            pq.Q(0, 1) | pq.Squeezing(r=0.5)

        simulator = pq.GaussianSimulator(d=5)
        result = simulator.execute(program)

    In the above example, the :class:`~piquasso.instructions.gates.Squeezing` gate
    is applied to modes `0, 1`.

    Note, that it is not necessarily required to specify any modes, if the context
    permits, e.g. when registering states.

    One could use the `all` keyword to indicate that the registered
    :class:`~piquasso.api.instruction.Instruction` mashould be applied to all modes,
    i.e.::

        with pq.Program() as program:
            pq.Q(0, 1) | pq.Squeezing(r=0.5)

            pq.Q(all) | pq.ParticleNumberMeasurement()

    or alternatively no parameters, i.e.::

        with pq.Program() as program:
            pq.Q(0, 1) | pq.Squeezing(r=0.5)

            pq.Q() | pq.ParticleNumberMeasurement()
    """

    def __init__(self, *modes: Union[int, Any]) -> None:
        """
        Args:
            modes (int):
                Variable length list of non-negative integers specifying the modes.

        Raises:
            InvalidModes:
                Raised if
                - the specified modes are not distinct;
                - negative integers were specified.
        """
        is_all = modes == (all,)

        if not is_all and any(mode < 0 for mode in modes):
            raise InvalidModes(
                f"Error registering modes: '{modes}' should be non-negative."
            )

        if not is_all and not self._is_distinct(modes):
            raise InvalidModes(
                f"Error registering modes: '{modes}' should be distinct."
            )

        self.modes: Tuple[int, ...] = modes if not is_all else tuple()

    def __or__(self, rhs: Union["Instruction", "Program"]) -> "Q":
        """Registers an `Instruction` or `Program` to the current program.

        If `rhs` is an `Instruction`, then it is appended to the current program's
        `instructions`.

        If `rhs` is a `Program`, then the current program's `instructions` is extended
        with the `instructions` of `rhs`.

        Args:
            rhs (Instruction or Program):
                An `Instruction` or a `Program` to be added to the current program.

        Returns:
            (Q): The current qumode.
        """

        if _context.current_program is None:
            raise PiquassoException(
                "The current program is None; therefore, it is not possible to register"
                " the parameters on the given modes.\n"
                "Please use `Q(...) | ...` only inside `with pq.Program() as program:` "
                "blocks, e.g.\n"
                "\n"
                "with pq.Program() as program:\n"
                "    pq.Q(0, 1) | pq.Squeezing(r=0.5)"
            )

        rhs._apply_to_program_on_register(_context.current_program, register=self)

        return self

    __ror__ = __or__

    @staticmethod
    def _is_distinct(iterable: Collection) -> bool:
        return len(iterable) == len(set(iterable))



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Gates can be characterized by a unitary operator :math:`U`, which evolves the
quantum state in the following manner

.. math::
    \rho' = U \rho U^\dagger

Gates with at most quadratic Hamiltonians are called linear gates. Evolution of the
ladder operators by linear gates could be expressed in the form

.. math::

    U \xi U^\dagger = S_{(c)} \xi + d,

where :math:`S_{(c)} \in \operatorname{Sp}(2d, \mathbb{R})`,
:math:`d \in \mathbb{C}^{2d}`,

.. math::
    \xi = \begin{bmatrix}
        a_1, \dots, a_d, a_1^\dagger, \dots, a_d^\dagger
    \end{bmatrix}^T,

where :math:`a_1^\dagger, ..., a_d^\dagger` and :math:`a_1, \dots, a_d` are the
creation and annihilation operators, respectively.

Most of the gates defined here are linear gates, which can be characterized by
:math:`S_{(c)}` and :math:`d`.
"""

import numpy as np

from scipy.optimize import root_scalar
from scipy.linalg import block_diag

from piquasso.api.instruction import Gate
from piquasso.api.errors import InvalidParameter

from piquasso._math.decompositions import takagi
from piquasso._math.linalg import is_square, is_symmetric, is_invertible
from piquasso._math.symplectic import complex_symplectic_form, is_symplectic

from piquasso.core import _mixins


class _BogoliubovTransformation(Gate):
    def __init__(
        self,
        *,
        params: dict = None,
        passive_block: np.ndarray = None,
        active_block: np.ndarray = None,
        displacement_vector: np.ndarray = None,
    ):
        params = params or {}

        super().__init__(
            params=params,
            extra_params=dict(
                passive_block=passive_block,
                active_block=active_block,
                displacement_vector=displacement_vector,
            ),
        )


class _ScalableBogoliubovTransformation(
    _BogoliubovTransformation,
    _mixins.ScalingMixin,
):
    ERROR_MESSAGE_TEMPLATE = (
        "The instruction {instruction} is not applicable to modes {modes} with the "
        "specified parameters."
    )

    def _autoscale(self) -> None:

        passive_block = self._extra_params["passive_block"]
        if passive_block is None or len(self.modes) == len(passive_block):
            pass
        elif len(passive_block) == 1:
            self._extra_params["passive_block"] = block_diag(
                *[passive_block] * len(self.modes)
            )
        else:
            raise InvalidParameter(
                self.ERROR_MESSAGE_TEMPLATE.format(instruction=self, modes=self.modes)
            )

        active_block = self._extra_params["active_block"]
        if active_block is None or len(self.modes) == len(active_block):
            pass
        elif len(active_block) == 1:
            self._extra_params["active_block"] = block_diag(
                *[active_block] * len(self.modes)
            )
        else:
            raise InvalidParameter(
                self.ERROR_MESSAGE_TEMPLATE.format(instruction=self, modes=self.modes)
            )

        displacement_vector = self._extra_params["displacement_vector"]
        if displacement_vector is None or len(self.modes) == len(displacement_vector):
            pass
        elif len(displacement_vector) == 1:
            self._extra_params["displacement_vector"] = np.array(
                [displacement_vector[0]] * len(self.modes),
                dtype=complex,
            )
        else:
            raise InvalidParameter(
                self.ERROR_MESSAGE_TEMPLATE.format(instruction=self, modes=self.modes)
            )


class Interferometer(_BogoliubovTransformation):
    r"""Applies a general interferometer gate.

    The general unitary operator can be written as

    .. math::
        U = \exp \left (
            i \sum_{i, j = 1}^d H_{i j} a_i^\dagger a_j
        \right ),

    where the parameter `U` and :math:`H` is related by

    .. math::
        U = \exp \left (
            i H \right
        ).

    The evolution of the ladder operators can be described by

    .. math::
        S_{(c)} = \begin{bmatrix}
            U & 0_{d \times d} \\
            0_{d \times d} & U^*
        \end{bmatrix},

    where :math:`U \in \mathbb{C}^{d \times d}` is a unitary matrix.

    Args:
        matrix (numpy.ndarray):
            The unitary interferometer matrix, corresponding to a
            passive transformation on the one-particle subspace.
    """

    def __init__(self, matrix: np.ndarray) -> None:
        if not is_square(matrix):
            raise InvalidParameter(
                "The interferometer matrix should be a square matrix."
            )

        super().__init__(params=dict(matrix=matrix), passive_block=matrix)


class Beamsplitter(_BogoliubovTransformation):
    r"""Applies a beamsplitter gate.

    The general unitary operator can be written as

    .. math::
        BS_{ij} (\theta, \phi) = \exp \left (
            \theta e^{i \phi} a^\dagger_i a_j
            - \theta e^{- i \phi} a^\dagger_j a_i
        \right ).

    The symplectic representation of the beamsplitter gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            t  & -r^* &    & \\
            r & t   &    & \\
               &     & t  & -r \\
               &     & r^* & t
        \end{bmatrix},

    where :math:`t = \cos(\theta)` and :math:`r = e^{i \phi} \sin(\theta)`.
    """

    def __init__(self, theta: float = 0.0, phi: float = np.pi / 4) -> None:
        """
        Args:
            phi (float):
                Phase angle of the beamsplitter.
                (defaults to :math:`\phi = \pi/2` that gives a symmetric beamsplitter)
            theta (float):
                The transmittivity angle of the beamsplitter.
                (defaults to :math:`\theta=\pi/4` that gives a 50-50 beamsplitter)
        """

        t = np.cos(theta)
        r = np.exp(1j * phi) * np.sin(theta)

        super().__init__(
            params=dict(
                theta=theta,
                phi=phi,
            ),
            passive_block=np.array(
                [
                    [t, -np.conj(r)],
                    [r, t],
                ]
            ),
        )


class Phaseshifter(_ScalableBogoliubovTransformation):
    r"""Applies a rotation or a phaseshifter gate.

    The unitary operator corresponding to the phaseshifter gate on the :math:`i`-th mode
    is

    .. math::
        R_i (\phi) = \exp \left (
            i \phi a_i^\dagger a_i
        \right ).

    The symplectic representation of the phaseshifter gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            e^{i \phi} & 0 \\
            0 & e^{- i \phi}
        \end{bmatrix}.
    """

    def __init__(self, phi: float) -> None:
        """
        Args:
            phi (float): The angle of the rotation.
        """

        super().__init__(
            params=dict(phi=phi), passive_block=np.diag(np.exp(1j * np.atleast_1d(phi)))
        )


class MachZehnder(_BogoliubovTransformation):
    r"""Mach-Zehnder interferometer.

    The Mach-Zehnder interferometer is equivalent to

    .. math::
        MZ(\phi_{int}, \phi_{ext}) =
            B(\pi/4, \pi/2) (R(\phi_{int}) \oplus \mathbb{1})
            B(\pi/4, \pi/2) (R(\phi_{ext}) \oplus \mathbb{1}).

    The symplectic representation of the Mach-Zehnder interferometer is

    .. math::
        S_{(c)} = \begin{bmatrix}
        e^{i \phi_{ext} } (e^{i \phi_{int} } - 1)   & i (e^{i \phi_{int} } + 1) & & \\
        i e^{i \phi_{ext} } (e^{i \phi_{int} } + 1) & 1 - e^{i \phi_{int} }     & & \\
        & & e^{-i \phi_{ext} } (e^{-i \phi_{int} } - 1) & -i (e^{-i \phi_{int} } + 1) \\
        & & -i e^{-i \phi_{ext} } (e^{-i \phi_{int} } + 1) & 1 - e^{-i \phi_{int} }
        \end{bmatrix},

    where :math:`\phi_{int}, \phi_{ext} \in \mathbb{R}`.
    """

    def __init__(self, int_: float, ext: float) -> None:
        """
        Args:
            int_ (float): The internal angle.
            ext (float): The external angle.
        """
        int_phase, ext_phase = np.exp(1j * np.array([int_, ext]))

        super().__init__(
            params=dict(int_=int_, ext=ext),
            passive_block=1
            / 2
            * np.array(
                [
                    [ext_phase * (int_phase - 1), 1j * (int_phase + 1)],
                    [1j * ext_phase * (int_phase + 1), 1 - int_phase],
                ]
            ),
        )


class Fourier(_ScalableBogoliubovTransformation):
    r"""Applies a Fourier gate. It simply transforms the quadratures as follows:

    .. math::
        -\hat{p} = \operatorname{\textit{F}} \hat{x} \operatorname{\textit{F}}^\dagger
            \\ \hat{x} = \operatorname{\textit{F}} \hat{p}
                \operatorname{\textit{F}}^\dagger

    The unitary operator corresponding to the Fourier gate on the :math:`i`-th mode is

    .. math::
        F_{i} = \exp \left (
            i \frac{\pi}{2} a_i^\dagger a_i
        \right ).

    The symplectic representation of the Fourier gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            i & 0  \\
            0 & -i \\
        \end{bmatrix}.

    Note:
        Corresponds to the :class:`Phaseshifter` gate with :math:`\phi = \pi/2`.
    """

    def __init__(self) -> None:
        super().__init__(passive_block=np.array([[1j]]))


class GaussianTransform(_BogoliubovTransformation):
    r"""Applies a Gaussian transformation gate.

    The symplectic representation of the Gaussian gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            P & A \\
            A^* & P^*
        \end{bmatrix},

    where :math:`P, A \in C^{d \times d}` fulfilling the relations

    .. math::
        P P^\dagger - A A^\dagger &= I_{d \times d} \\
        P A^T &= A P^T
    """

    def __init__(self, passive: np.ndarray, active: np.ndarray) -> None:
        """
        Args:
            passive (numpy.ndarray):
                The passive submatrix of the symplectic matrix corresponding to the
                transformation.
            active (numpy.ndarray):
                The active submatrix of the symplectic matrix corresponding to the
                transformation.

        Raises:
            InvalidParameters: Raised if the parameters do not form a symplectic matrix.
        """
        if not is_symplectic(
            np.block([[passive, active], [active.conj(), passive.conj()]]),
            form_func=complex_symplectic_form,
        ):
            raise InvalidParameter(
                "The input parameters for instruction 'GaussianTransform' do not form "
                "a symplectic matrix."
            )

        super().__init__(
            params=dict(passive=passive, active=active),
            passive_block=passive,
            active_block=active,
        )


class Squeezing(_ScalableBogoliubovTransformation):
    r"""Applies the squeezing gate.

    The unitary operator corresponding to the squeezing gate is

    .. math::
        S_{i} (z) = \exp \left (
            \frac{1}{2}(z^* a_i^2 - z a_i^{\dagger 2} )
        \right ).

    The symplectic representation of the squeezing gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            \cosh r & - e^{i \phi} \sinh r \\
            - e^{- i \phi} \sinh r & \cosh r
        \end{bmatrix}.

    The unitary squeezing operator is

    .. math::
        S(z) = \exp \left (
            \frac{1}{2}(z^* a_i^2 - z a_i^{\dagger 2})
        \right ),

    where :math:`z = re^{i\phi}`.
    """

    def __init__(self, r: float, phi: float = 0.0) -> None:
        """
        Args:
            r (float): The amplitude of the squeezing instruction.
            phi (float): The squeezing angle.
        """
        super().__init__(
            params=dict(r=r, phi=phi),
            passive_block=np.diag(np.atleast_1d(np.cosh(r))),
            active_block=np.diag(
                -np.atleast_1d(np.sinh(r)) * np.exp(1j * np.atleast_1d(phi))
            ),
        )


class QuadraticPhase(_ScalableBogoliubovTransformation):
    r"""Applies the quadratic phase instruction to the state.

    The unitary operator corresponding to the Fourier gate is

    .. math::
        QP_{i} (s) = \exp \left (
            i \frac{s}{2 \hbar} x_i^2
        \right ).

    The symplectic representation of the quadratic phase gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            1 + i \frac{s}{2} & i \frac{s}{2} \\
            -i \frac{s}{2} & 1 - i \frac{s}{2}
        \end{bmatrix}.

    """

    def __init__(self, s: float) -> None:
        super().__init__(
            params=dict(s=s),
            passive_block=np.diag(1 + np.atleast_1d(s) / 2 * 1j),
            active_block=np.diag(np.atleast_1d(s) / 2 * 1j),
        )


class Squeezing2(_BogoliubovTransformation):
    r"""Applies the two-mode squeezing gate to the state.

    The unitary operator corresponding to the two-mode squeezing gate is

    .. math::
        S_{ij} (z) = \exp \left (
            \frac{1}{2}(z^* a_i a_j - z a_i^\dagger a_j^\dagger )
        \right ).

    The symplectic representation of the two-mode squeezing gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            \cosh r & 0 & 0 & e^{i \phi} \sinh r \\
            0 & \cosh r & e^{i \phi} \sinh r & 0 \\
            0 & e^{- i \phi} \sinh r & \cosh r & 0 \\
            e^{- i \phi} \sinh r & 0 & 0 & \cosh r
        \end{bmatrix}.
    """

    def __init__(self, r: float, phi: float = 0.0) -> None:
        """
        Args:
            r (float): The amplitude of the squeezing instruction.
            phi (float): The squeezing angle.
        """
        super().__init__(
            params=dict(r=r, phi=phi),
            passive_block=np.array(
                [
                    [np.cosh(r), 0],
                    [0, np.cosh(r)],
                ]
            ),
            active_block=np.array(
                [
                    [0, np.sinh(r) * np.exp(1j * phi)],
                    [np.sinh(r) * np.exp(1j * phi), 0],
                ]
            ),
        )


class ControlledX(_BogoliubovTransformation):
    r"""Applies the controlled X gate to the state.

    The unitary operator corresponding to the controlled X gate is

    .. math::
        CX_{ij} (s) = \exp \left (
            -i \frac{s}{\hbar} x_i p_j
        \right ).

    The symplectic representation of the controlled X gate is

    .. math::
        S_{(z)} = \begin{bmatrix}
            1            & - \frac{s}{2} & 0           & \frac{s}{2} \\
            \frac{s}{2} & 1              & \frac{s}{2} & 0            \\
            0            & \frac{s}{2} & 1            & - \frac{s}{2} \\
            \frac{s}{2} & 0            & \frac{s}{2} & 1
        \end{bmatrix}.
    """

    def __init__(self, s: float):
        super().__init__(
            params=dict(s=s),
            passive_block=np.array(
                [
                    [1, -s / 2],
                    [s / 2, 1],
                ]
            ),
            active_block=np.array(
                [
                    [0, s / 2],
                    [s / 2, 0],
                ]
            ),
        )


class ControlledZ(_BogoliubovTransformation):
    r"""Applies the controlled Z gate to the state.

    The unitary operator corresponding to the controlled Z gate is

    .. math::
        CZ_{ij} (s) = \exp \left (
            -i \frac{s}{\hbar} x_i x_j
        \right ).

    The symplectic representation of the controlled Z gate is

    .. math::
        S_{(z)} = \begin{bmatrix}
            1            & i \frac{s}{2} & 0            & i \frac{s}{2}    \\
            i \frac{s}{2} & 1              & i \frac{s}{2} & 0             \\
            0            & -i \frac{s}{2} & 1            & - i \frac{s}{2} \\
            -i \frac{s}{2} & 0            & -i \frac{s}{2} & 1
        \end{bmatrix}.
    """

    def __init__(self, s: float):
        super().__init__(
            params=dict(s=s),
            passive_block=np.array(
                [
                    [1, 1j * (s / 2)],
                    [1j * (s / 2), 1],
                ]
            ),
            active_block=np.array(
                [
                    [0, 1j * (s / 2)],
                    [1j * (s / 2), 0],
                ]
            ),
        )


class Displacement(_ScalableBogoliubovTransformation):
    r"""Phase space displacement instruction.

    Evolves the ladder operators by

    .. math::
        \xi \mapsto \xi + \begin{bmatrix}
            \alpha \\
            \alpha^*
        \end{bmatrix},

    where :math:`\alpha \in \mathbb{C}^{d \times d}`.

    One must either specify :math:`\alpha` only, or the combination of :math:`r` and
    :math:`\phi`.
    When :math:`r` and :math:`\phi` are the given parameters, :math:`\alpha` is
    calculated via:

    .. math::
        \alpha = r e^{i\phi}
    """

    def __init__(
        self, *, alpha: complex = None, r: float = None, phi: float = None
    ) -> None:
        """
        Args:
            alpha (complex): The displacement.
            r (float): The displacement magnitude.
            phi (float): The displacement angle.
        """
        alpha_: np.ndarray

        if alpha is not None and r is None and phi is None:
            params = dict(alpha=alpha)
            alpha_ = np.atleast_1d(alpha)
        elif alpha is None and r is not None and phi is not None:
            params = dict(r=r, phi=phi)
            alpha_ = np.atleast_1d(r) * np.exp(1j * np.atleast_1d(phi))
        else:
            raise InvalidParameter(
                "Either specify 'alpha' only, or the combination of 'r' and 'phi': "
                f"alpha={alpha}, r={r}, phi={phi}."
            )

        super().__init__(params=params, displacement_vector=alpha_)


class PositionDisplacement(_ScalableBogoliubovTransformation):
    r"""Position displacement gate. It affects only the :math:`\hat{x}` quadrature.

    Note:
        The specified displacement is automatically scaled by :math:`\sqrt{2 \hbar}`.
    """

    def __init__(self, x: float) -> None:
        """
        Args:
            x (float): The position displacement.
        """
        super().__init__(
            params=dict(x=x),
            displacement_vector=np.atleast_1d(x),
        )


class MomentumDisplacement(_ScalableBogoliubovTransformation):
    r"""Momentum displacement gate. It only affects the :math:`\hat{p}` quadrature.

    Note:
        The specified displacement is automatically scaled by :math:`\sqrt{2 \hbar}`.
    """

    def __init__(self, p: float) -> None:
        """
        Args:
            p (float): The momentum displacement.
        """
        super().__init__(
            params=dict(p=p),
            displacement_vector=1j * np.atleast_1d(p),
        )


class _ScalableFockGates(Gate, _mixins.ScalingMixin):
    def __init__(
        self, *, params: dict = None, gamma: np.ndarray = None, xi: np.ndarray = None
    ):
        params = params or {}
        super().__init__(params=params, extra_params=dict(gamma=gamma, xi=xi))

    ERROR_MESSAGE_TEMPLATE = (
        "The instruction {instruction} is not applicable to modes {modes} with the "
        "specified parameters."
    )

    def _autoscale(self) -> None:
        gamma = self._extra_params["gamma"]
        xi = self._extra_params["xi"]
        if gamma is not None and len(gamma) == len(self.modes):
            pass
        elif gamma is None:
            pass
        elif len(gamma) == 1:
            self._extra_params["gamma"] = np.array([gamma] * len(self.modes))
        else:
            raise InvalidParameter(
                self.ERROR_MESSAGE_TEMPLATE.format(instruction=self, modes=self.modes)
            )

        if xi is not None and len(xi) == len(self.modes):
            pass
        elif xi is None:
            pass
        elif len(xi) == 1:
            self._extra_params["xi"] = np.array([xi] * len(self.modes))
        else:
            raise InvalidParameter(
                self.ERROR_MESSAGE_TEMPLATE.format(instruction=self, modes=self.modes)
            )


class CubicPhase(_ScalableFockGates):
    r"""Cubic Phase gate.

    The definition of the Cubic Phase gate is

    .. math::
        \operatorname{CP}(\gamma) = e^{i \hat{x}^3 \frac{\gamma}{3 \hbar}}

    The Cubic Phase gate transforms the annihilation operator as

    .. math::
        \operatorname{CP}^\dagger(\gamma) \hat{a} \operatorname{CP}(\gamma) = \hat{a}
            + i\frac{\gamma(\hat{a} +\hat{a}^\dagger)^2}{2\sqrt{2/\hbar}}

    It transforms the :math:`\hat{p}` quadrature as follows:

    .. math::
        \operatorname{CP}^\dagger(\gamma) \hat{p} \operatorname{CP}(\gamma) =
            \hat{p} + \gamma \hat{x}^2.

    Note:
        This is a non-linear gate, therefore it couldn't be used with
        :class:`~piquasso._backends.gaussian.state.GaussianState`.
        Using this gate requires a high cutoff to make the more accurate simulation.
    """

    def __init__(self, gamma: float) -> None:
        """
        Args:
            gamma (float): The Cubic Phase parameter.
        """
        super().__init__(params=dict(gamma=gamma), gamma=np.atleast_1d(gamma))


class Kerr(_ScalableFockGates):
    r"""Kerr gate.

    The definition of the Kerr gate is

    .. math::
        K_i (\xi) = \exp \left (
            i \xi n_i n_i
        \right ).

    The Kerr gate transforms the annihilation operator as

    .. math::
        K(\xi) a K(\xi) = a \exp(- i \xi (1 + 2 n)).

    Note:
        This is a non-linear gate, therefore it couldn't be used with
        :class:`~piquasso._backends.gaussian.state.GaussianState`.
    """

    def __init__(self, xi: float) -> None:
        """
        Args:
            xi (float): The magnitude of the Kerr nonlinear term.
        """
        super().__init__(params=dict(xi=xi), xi=np.atleast_1d(xi))


class CrossKerr(Gate):
    r"""Cross-Kerr gate.

    The definition of the Cross-Kerr gate is

    .. math::
        CK_{ij} (\xi) = \exp \left (
            i \xi n_i n_j
        \right )

    The Cross-Kerr gate transforms the annihilation operators as

    .. math::
        CK_{ij} (\xi) a_i CK_{ij} (\xi) &= a_i \exp(- i \xi n_j) \\
        CK_{ij} (\xi) a_j CK_{ij} (\xi) &= a_j \exp(- i \xi n_i)


    Note:
        This is a non-linear gate, therefore it couldn't be used with
        :class:`~piquasso._backends.gaussian.state.GaussianState`.
    """

    def __init__(self, xi: float) -> None:
        """
        Args:
            xi (float): The magnitude of the Cross-Kerr nonlinear term.
        """
        super().__init__(params=dict(xi=xi))


class Graph(Gate):
    r"""Applies a graph given its adjacency matrix, see
    `this article <https://arxiv.org/pdf/1612.01199.pdf>`_ for more details.
    It decomposes the adjacency matrix of a graph into single mode squeezers and
    interferometers.
    """

    def __init__(
        self, adjacency_matrix: np.ndarray, mean_photon_number: float = 1.0
    ) -> None:
        r"""
        Args:
            adjacency_matrix (numpy.ndarray): A symmetric matrix with a size of :math:`N
                \times N` and it can be real or complex.
            mean_photon_number (float, optional): The mean photon number :math:`\bar{n}`
                for a mode. Defaults to :math:`1.0`.
        Raises:
            InvalidParameter:
                If the adjacency matrix is not invertible or not symmetric.
        """
        self.adjacency_matrix = adjacency_matrix

        if not is_invertible(adjacency_matrix):
            raise InvalidParameter("The adjacency matrix is not invertible.")

        if not is_symmetric(adjacency_matrix):
            raise InvalidParameter("The adjacency matrix should be symmetric.")

        singular_values, unitary = takagi(adjacency_matrix)

        scaling = self._get_scaling(singular_values, mean_photon_number)

        squeezing_parameters = np.arctanh(scaling * singular_values)

        # TODO: find a better solution for these.
        squeezing = GaussianTransform(
            passive=np.diag(np.cosh(squeezing_parameters)),
            active=np.diag(np.sinh(squeezing_parameters)),
        )

        interferometer = Interferometer(unitary)

        super().__init__(
            params=dict(
                adjacency_matrix=adjacency_matrix,
                mean_photon_number=mean_photon_number,
            ),
            extra_params=dict(
                squeezing=squeezing,
                interferometer=interferometer,
            ),
        )

    def _get_scaling(
        self, singular_values: np.ndarray, mean_photon_number: float
    ) -> float:
        r"""
        For a squeezed state :math:`rho` the mean photon number is calculated by

        .. math::
            \langle n \rangle_\rho = \sum_{i = 0}^d \mathrm{sinh}(r_i)^2

        where :math:`r_i = \mathrm{arctan}(s_i)`, where :math:`s_i` are the singular
        values of the adjacency matrix.
        """

        def mean_photon_number_equation(scaling: float) -> float:
            return (
                sum(
                    (scaling * singular_value) ** 2
                    / (1 - (scaling * singular_value) ** 2)
                    for singular_value in singular_values
                )
                / len(singular_values)
                - mean_photon_number
            )

        def mean_photon_number_gradient(scaling: float) -> float:
            return (2.0 / scaling) * np.sum(
                (singular_values * scaling / (1 - (singular_values * scaling) ** 2))
                ** 2
            )

        lower_bound = 0.0

        tolerance = 1e-10  # Needed to avoid zero division.

        upper_bound = 1.0 / (max(singular_values) + tolerance)

        result = root_scalar(
            mean_photon_number_equation,
            fprime=mean_photon_number_gradient,
            x0=(lower_bound - upper_bound) / 2.0,
            bracket=(lower_bound, upper_bound),
        )

        if not result.converged:
            raise InvalidParameter(
                f"No scaling found for adjacency matrix: {self.adjacency_matrix}."
            )

        return result.root



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, List

import itertools
import scipy
import random
import numpy as np

from itertools import repeat
from functools import lru_cache

from scipy.special import factorial

from .state import GaussianState
from .probabilities import calculate_click_probability

from piquasso.api.result import Result
from piquasso.api.instruction import Instruction
from piquasso.api.errors import InvalidInstruction

from piquasso._math.linalg import reduce_
from piquasso._math.indices import get_operator_index, get_auxiliary_operator_index
from piquasso._math.decompositions import decompose_to_pure_and_mixed


def passive_linear(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    modes = instruction.modes
    T: np.ndarray = instruction._all_params["passive_block"]

    state._m[(modes,)] = (
        T
        @ state._m[
            modes,
        ]
    )

    _apply_passive_linear_to_C_and_G(state, T, modes=modes)

    return Result(state=state)


def _apply_passive_linear_to_C_and_G(
    state: GaussianState, T: np.ndarray, modes: Tuple[int, ...]
) -> None:
    index = get_operator_index(modes)

    state._C[index] = T.conjugate() @ state._C[index] @ T.transpose()
    state._G[index] = T @ state._G[index] @ T.transpose()

    auxiliary_modes = state._get_auxiliary_modes(modes)

    if len(auxiliary_modes) != 0:
        _apply_passive_linear_to_auxiliary_modes(state, T, modes, auxiliary_modes)


def _apply_passive_linear_to_auxiliary_modes(
    state: GaussianState,
    T: np.ndarray,
    modes: Tuple[int, ...],
    auxiliary_modes: Tuple[int, ...],
) -> None:
    auxiliary_index = get_auxiliary_operator_index(modes, auxiliary_modes)

    state._C[auxiliary_index] = T.conjugate() @ state._C[auxiliary_index]
    state._G[auxiliary_index] = T @ state._G[auxiliary_index]

    state._C[:, modes] = np.conj(state._C[modes, :]).transpose()
    state._G[:, modes] = state._G[modes, :].transpose()


def linear(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    modes = instruction.modes
    passive_block: np.ndarray = instruction._all_params["passive_block"]
    active_block: np.ndarray = instruction._all_params["active_block"]

    state._m[(modes,)] = passive_block @ state._m[(modes,)] + active_block @ np.conj(
        state._m[
            modes,
        ]
    )

    _apply_linear_to_C_and_G(state, passive_block, active_block, modes)

    return Result(state=state)


def _apply_linear_to_C_and_G(
    state: GaussianState, P: np.ndarray, A: np.ndarray, modes: Tuple[int, ...]
) -> None:
    index = get_operator_index(modes)

    original_C = state._C[index]
    original_G = state._G[index]

    state._G[index] = (
        P @ original_G @ P.transpose()
        + A @ original_G.conjugate().transpose() @ A.transpose()
        + P @ (original_C.transpose() + np.identity(len(modes))) @ A.transpose()
        + A @ original_C @ P.transpose()
    )

    state._C[index] = (
        P.conjugate() @ original_C @ P.transpose()
        + A.conjugate()
        @ (original_C.transpose() + np.identity(len(modes)))
        @ A.transpose()
        + P.conjugate() @ original_G.conjugate().transpose() @ A.transpose()
        + A.conjugate() @ original_G @ P.transpose()
    )

    auxiliary_modes = state._get_auxiliary_modes(modes)

    if len(auxiliary_modes) != 0:
        _apply_linear_to_auxiliary_modes(state, P, A, modes, auxiliary_modes)


def _apply_linear_to_auxiliary_modes(
    state: GaussianState,
    P: np.ndarray,
    A: np.ndarray,
    modes: Tuple[int, ...],
    auxiliary_modes: Tuple[int, ...],
) -> None:
    auxiliary_index = get_auxiliary_operator_index(modes, auxiliary_modes)

    auxiliary_C = state._C[auxiliary_index]
    auxiliary_G = state._G[auxiliary_index]

    state._C[auxiliary_index] = (
        P.conjugate() @ auxiliary_C + A.conjugate() @ auxiliary_G
    )

    state._G[auxiliary_index] = P @ auxiliary_G + A @ auxiliary_C

    state._C[:, modes] = state._C[modes, :].conjugate().transpose()
    state._G[:, modes] = state._G[modes, :].transpose()


def displacement(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    modes = instruction.modes
    displacement_vector: np.ndarray = instruction._all_params["displacement_vector"]

    state._m[
        modes,
    ] += displacement_vector

    return Result(state=state)


def generaldyne_measurement(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    modes = instruction.modes
    detection_covariance = instruction._all_params["detection_covariance"]

    samples = _get_generaldyne_samples(
        state,
        modes,
        shots,
        detection_covariance,
    )

    # NOTE: We choose the last sample for multiple shots.
    sample = samples[-1]

    d = len(state)

    auxiliary_modes = state._get_auxiliary_modes(modes)
    outer_indices = _map_modes_to_xpxp_indices(auxiliary_modes)

    evolved_state = _get_generaldyne_evolved_state(
        state,
        sample,
        modes,
        detection_covariance,
    )

    evolved_mean = np.zeros(2 * d)
    evolved_mean[outer_indices] = evolved_state.xpxp_mean_vector

    evolved_cov = np.identity(2 * d) * state._config.hbar
    evolved_cov[
        np.ix_(outer_indices, outer_indices)
    ] = evolved_state.xpxp_covariance_matrix

    state.xpxp_mean_vector = evolved_mean
    state.xpxp_covariance_matrix = evolved_cov

    return Result(state=state, samples=list(map(tuple, list(samples))))


def _get_generaldyne_samples(state, modes, shots, detection_covariance):
    indices = _map_modes_to_xpxp_indices(modes)

    full_detection_covariance = state._config.hbar * scipy.linalg.block_diag(
        *[detection_covariance] * len(modes)
    )

    mean = state.xpxp_mean_vector[indices]

    cov = (
        state.xpxp_covariance_matrix[np.ix_(indices, indices)]
        + full_detection_covariance
    )

    # HACK: We need tol=1e-7 to avoid Numpy warnings at homodyne detection with
    # squeezed detection covariance. Numpy warns
    # 'covariance is not positive-semidefinite.', but it definitely is. In the SVG
    # decomposition (which numpy uses for the multivariate normal distribution)
    # the U^T and V matrices should equal, but our covariance might contain too
    # large values leading to inequality, resulting in warning.
    #
    # We might be better of setting `check_valid='ignore'` and verifying
    # positive-definiteness for ourselves.
    return state._config.rng.multivariate_normal(
        mean=mean,
        cov=cov,
        size=shots,
        tol=1e-7,
    )


def _get_generaldyne_evolved_state(state, sample, modes, detection_covariance):
    full_detection_covariance = state._config.hbar * scipy.linalg.block_diag(
        *[detection_covariance] * len(modes)
    )

    mean = state.xpxp_mean_vector
    cov = state.xpxp_covariance_matrix

    indices = _map_modes_to_xpxp_indices(modes)
    outer_indices = np.delete(np.arange(2 * state.d), indices)

    mean_measured = mean[indices]
    mean_outer = mean[outer_indices]

    cov_measured = cov[np.ix_(indices, indices)]
    cov_outer = cov[np.ix_(outer_indices, outer_indices)]
    cov_correlation = cov[np.ix_(outer_indices, indices)]

    evolved_cov_outer = (
        cov_outer
        - cov_correlation
        @ np.linalg.inv(cov_measured + full_detection_covariance)
        @ cov_correlation.transpose()
    )

    evolved_r_A = mean_outer + cov_correlation @ np.linalg.inv(
        cov_measured + full_detection_covariance
    ) @ (sample - mean_measured)

    state = GaussianState(d=len(evolved_r_A) // 2, config=state._config)

    state.xpxp_covariance_matrix = evolved_cov_outer
    state.xpxp_mean_vector = evolved_r_A

    return state


def _map_modes_to_xpxp_indices(modes):
    indices = []

    for mode in modes:
        indices.extend([2 * mode, 2 * mode + 1])

    return indices


def particle_number_measurement(
    state: GaussianState,
    instruction: Instruction,
    shots: int,
) -> Result:

    samples = _get_particle_number_measurement_samples(state, instruction, shots)

    return Result(state=state, samples=samples)


def _get_particle_number_measurement_samples(
    state: GaussianState,
    instruction: Instruction,
    shots: int,
) -> List[Tuple[int, ...]]:

    modes: Tuple[int, ...] = instruction.modes

    reduced_state = state.reduced(modes)

    reduced_modes = tuple(range(len(modes)))

    pure_covariance, mixed_contribution = decompose_to_pure_and_mixed(
        reduced_state.xxpp_covariance_matrix,
        hbar=state._config.hbar,
    )
    pure_state = GaussianState(len(reduced_state), config=state._config)
    pure_state.xxpp_covariance_matrix = pure_covariance

    heterodyne_detection_covariance = np.identity(2)

    samples = []

    for _ in itertools.repeat(None, shots):
        pure_state.xxpp_mean_vector = state._config.rng.multivariate_normal(
            reduced_state.xxpp_mean_vector, mixed_contribution
        )

        heterodyne_sample = _get_generaldyne_samples(
            state=pure_state,
            modes=reduced_modes,
            shots=1,
            detection_covariance=heterodyne_detection_covariance,
        )[0]

        sample: Tuple[int, ...] = tuple()

        heterodyne_measured_modes = reduced_modes

        for _ in itertools.repeat(None, len(reduced_modes)):
            heterodyne_sample = heterodyne_sample[2:]
            heterodyne_measured_modes = heterodyne_measured_modes[1:]

            evolved_state = _get_generaldyne_evolved_state(
                pure_state,
                heterodyne_sample,
                heterodyne_measured_modes,
                heterodyne_detection_covariance,
            )

            choice = _get_particle_number_choice(
                evolved_state,
                sample,
            )

            sample = sample + (choice,)

        samples.append(sample)

    return samples


def _get_particle_number_choice(
    state: GaussianState,
    previous_sample: Tuple[int, ...],
) -> int:
    r"""
    The original equations are

    .. math::
        A = X - X Q^{-1} = \begin{bmatrix}
            B & 0 \\
            0 & B^* \\
        \end{bmatrix}

    with

    .. math::
        X = \begin{bmatrix}
            0 & \mathbb{I} \\
            \mathbb{I} & 0 \\
        \end{bmatrix}

    But then one could write

    .. math::
        B = - Q^{-1}[d:, :d]

    Everything can be rewritten using :math:`B` in the following manner:

    ..math::
        \gamma = \alpha^* - \alpha B \\
        p_{vac} = \exp
            \left (
                - \Re(\gamma \alpha)
            \right )
            \sqrt{
                \operatorname{det}(\mathbb{I} - B^* B)
            } \\
        p_{S} = p_{vac} \frac{
            | \operatorname{lhaf}(\operatorname{filldiag}(B_S, \gamma_S))
        }{
            S!
        }
    """

    d = len(state)

    B = -np.linalg.inv(state.Q_matrix)[d:, :d]
    alpha = state.complex_displacement[:d]

    gamma = alpha.conj() - alpha @ B

    weights: np.ndarray = np.array([])

    possible_choices = tuple(range(state._config.measurement_cutoff))

    for n in possible_choices:
        occupation_numbers = previous_sample + (n,)

        B_reduced = reduce_(B, reduce_on=occupation_numbers)
        gamma_reduced = reduce_(gamma, reduce_on=occupation_numbers)

        np.fill_diagonal(B_reduced, gamma_reduced)

        weight = abs(state._config.loop_hafnian_function(B_reduced)) ** 2 / factorial(n)
        weights = np.append(weights, weight)

    weights /= np.sum(weights)

    return np.random.choice(possible_choices, p=weights)


def threshold_measurement(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    """
    NOTE: The same logic is used here, as for the particle number measurement.
    However, at threshold measurement there is no sense of measurement cutoff,
    therefore it is set to 2 to make the logic sensible in this case as well.

    Also note, that one could speed up this calculation by not calculating the
    probability of clicks (i.e. 1 as sample), and argue that the probability of a
    click is equal to one minus the probability of no click.
    """
    if state._config.use_torontonian:
        samples = _generate_threshold_samples_using_torontonian(
            state, instruction, shots
        )
    else:
        samples = _generate_threshold_samples_using_hafnian(state, instruction, shots)

    return Result(state=state, samples=samples)


def _generate_threshold_samples_using_torontonian(
    state,
    instruction,
    shots,
):
    if not np.allclose(state.xpxp_mean_vector, np.zeros_like(state.xpxp_mean_vector)):
        raise NotImplementedError(
            "Threshold measurement for displaced states are not supported: "
            f"xpxp_mean_vector={state.xpxp_mean_vector}"
        )

    modes = instruction.modes

    @lru_cache(state._config.cache_size)
    def get_probability(
        subspace_modes: Tuple[int, ...], occupation_numbers: Tuple[int, ...]
    ) -> float:
        reduced_state = state.reduced(subspace_modes)

        return calculate_click_probability(
            reduced_state.xxpp_covariance_matrix,
            occupation_numbers,
            state._config.hbar,
        )

    samples = []

    for _ in repeat(None, shots):
        sample: List[int] = []

        previous_probability = 1.0

        for mode_index in range(1, len(modes) + 1):
            subspace_modes = tuple(modes[:mode_index])

            occupation_numbers = tuple(sample + [0])

            probability = get_probability(
                subspace_modes=subspace_modes, occupation_numbers=occupation_numbers
            )
            conditional_probability = probability / previous_probability

            choice: int
            guess = random.uniform(0.0, 1.0)

            if guess < conditional_probability:
                choice = 0
                previous_probability *= conditional_probability
            else:
                choice = 1
                previous_probability *= 1 - conditional_probability

            sample.append(choice)

        samples.append(tuple(sample))

    return samples


def _generate_threshold_samples_using_hafnian(state, instruction, shots):
    samples = _get_particle_number_measurement_samples(state, instruction, shots)

    threshold_samples = []

    for sample in samples:
        threshold_samples.append(
            [1 if photon_number else 0 for photon_number in sample]
        )

    return threshold_samples


def homodyne_measurement(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    phi = instruction._all_params["phi"]

    instruction_copy: Instruction = instruction.copy()  # type: ignore

    phaseshift = np.identity(len(instruction.modes)) * np.exp(-1j * phi)
    instruction_copy._extra_params["passive_block"] = phaseshift

    result = passive_linear(state, instruction_copy, shots)

    result = generaldyne_measurement(result.state, instruction, shots)

    return Result(state=state, samples=result.samples)


def vacuum(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    state.reset()

    return Result(state=state)


def mean(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    state.xpxp_mean_vector = instruction._all_params["mean"] * np.sqrt(
        state._config.hbar
    )

    return Result(state=state)


def covariance(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    state.xpxp_covariance_matrix = instruction._all_params["cov"] * state._config.hbar

    return Result(state=state)


def graph(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    """
    TODO: Find a better solution for multiple operations.
    """
    instruction._all_params["squeezing"].modes = instruction.modes
    instruction._all_params["interferometer"].modes = instruction.modes

    result = linear(state, instruction._all_params["squeezing"], shots)

    return passive_linear(
        result.state, instruction._all_params["interferometer"], shots
    )


def deterministic_gaussian_channel(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    X = instruction._all_params["X"]
    Y = instruction._all_params["Y"] * state._config.hbar
    modes = instruction.modes

    if len(X) != 2 * len(modes) or len(Y) != 2 * len(modes):
        raise InvalidInstruction(
            f"The instruction should be specified for '{len(modes)}' modes: "
            f"instruction={instruction}"
        )

    indices = _map_modes_to_xpxp_indices(modes)
    matrix_indices = np.ix_(indices, indices)

    mean_vector = state.xpxp_mean_vector
    covariance_matrix = state.xpxp_covariance_matrix

    embedded_X = np.identity(len(mean_vector), dtype=complex)
    embedded_X[matrix_indices] = X

    embedded_Y = np.zeros_like(embedded_X)
    embedded_Y[matrix_indices] = Y

    state.xpxp_mean_vector = embedded_X @ mean_vector
    state.xpxp_covariance_matrix = (
        embedded_X @ covariance_matrix @ embedded_X.T + embedded_Y
    )

    return Result(state=state)



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, List
import numpy as np

from piquasso._math.combinatorics import partitions
from piquasso._math.fock import symmetric_subspace_cardinality
from piquasso._math.linalg import is_unitary
from piquasso._math.validations import all_natural

from piquasso.api.config import Config
from piquasso.api.errors import InvalidState
from piquasso.api.state import State
from piquasso.api.errors import PiquassoException

from theboss.distribution_calculators.bs_distribution_calculator_with_fixed_losses import (  # noqa: E501
    BSDistributionCalculatorWithFixedLosses,
    BosonSamplingExperimentConfiguration,
)

from theboss.boson_sampling_utilities.permanent_calculators.ryser_guan_permanent_calculator import (  # noqa: E501
    RyserGuanPermanentCalculator,
)


class SamplingState(State):
    def __init__(self, d: int, config: Config = None) -> None:
        super().__init__(config=config)

        self.initial_state: np.ndarray = np.zeros((d,), dtype=int)
        self.interferometer: np.ndarray = np.diag(np.ones(d, dtype=complex))

        self.is_lossy = False

    def validate(self) -> None:
        """Validates the currect state.

        Raises:
            InvalidState: If the interferometer matrix is non-unitary.
        """

        if not is_unitary(self.interferometer):
            raise InvalidState("The interferometer matrix is not unitary.")

        if not all_natural(self.initial_state):
            raise InvalidState(
                f"Invalid initial state: initial_state={self.initial_state}"
            )

    @property
    def d(self) -> int:
        r"""The number of modes, on which the state is defined."""
        return len(self.initial_state)

    @property
    def particle_number(self) -> int:
        r"""The number of particles in the system."""
        return sum(self.initial_state)

    def get_particle_detection_probability(
        self, occupation_number: Tuple[int, ...]
    ) -> float:
        if len(occupation_number) != self.d:
            raise PiquassoException(
                f"The specified occupation number should have length '{self.d}': "
                f"occupation_number='{occupation_number}'."
            )

        number_of_particles = sum(occupation_number)

        if number_of_particles != self.particle_number:
            return 0.0

        basis = partitions(self.d, number_of_particles)

        index = basis.index(occupation_number)

        subspace_probabilities = self._get_fock_probabilities_on_subspace()

        return subspace_probabilities[index]

    @property
    def fock_probabilities(self) -> np.ndarray:
        # TODO: All the `fock_probabilities` properties return a list according to the
        # `cutoff` specified in `config`. However, here it does not make sense to adhere
        # to that...

        cutoff = self.particle_number + 1

        probabilities = []

        for particle_number in range(cutoff):

            if particle_number == self.particle_number:
                subspace_probabilities = self._get_fock_probabilities_on_subspace()
            else:
                cardinality = symmetric_subspace_cardinality(
                    d=self.d, n=particle_number
                )
                subspace_probabilities = [0.0] * cardinality

            probabilities.extend(subspace_probabilities)

        return np.array(probabilities)

    def _get_fock_probabilities_on_subspace(self) -> List[float]:
        """
        The order of the returned Fock states is lexicographic, according to
        `theboss.boson_sampling_utilities.boson_sampling_utilities
        .generate_possible_outputs`
        """
        permanent_calculator = RyserGuanPermanentCalculator(self.interferometer)
        config = BosonSamplingExperimentConfiguration(
            interferometer_matrix=self.interferometer,
            initial_state=np.asarray(self.initial_state),
            number_of_modes=self.d,
            initial_number_of_particles=self.particle_number,
            number_of_particles_lost=0,
            number_of_particles_left=self.particle_number,
        )
        distribution_calculator = BSDistributionCalculatorWithFixedLosses(
            config, permanent_calculator
        )
        return distribution_calculator.calculate_distribution()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SamplingState):
            return False

        return (
            np.allclose(self.initial_state, other.initial_state)
            and np.allclose(self.interferometer, other.interferometer)
            and self.is_lossy == other.is_lossy
        )



#
# Copyright 2021 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import numpy as np
from piquasso._backends.sampling.state import SamplingState

from piquasso._math.validations import all_natural

from piquasso.api.errors import InvalidState
from piquasso.api.result import Result
from piquasso.api.instruction import Instruction

from theboss.boson_sampling_simulator import BosonSamplingSimulator

# The fastest implemented permanent calculator is currently Ryser-Guan
from theboss.boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_interface import (  # noqa: E501
    BSPermanentCalculatorInterface,
)
from theboss.boson_sampling_utilities.permanent_calculators.ryser_guan_permanent_calculator import (  # noqa: E501
    RyserGuanPermanentCalculator,
)

# Fastest boson sampling algorithm generalized for bunched states
from theboss.simulation_strategies.generalized_cliffords_simulation_strategy import (
    GeneralizedCliffordsSimulationStrategy,
)
from theboss.simulation_strategies.generalized_cliffords_uniform_losses_simulation_strategy import (  # noqa: E501
    GeneralizedCliffordsUniformLossesSimulationStrategy,
)

# Fastest BS algorithm generalized for bunched states, but with lossy network
from theboss.simulation_strategies.lossy_networks_generalized_cliffords_simulation_strategy import (  # noqa: E501
    LossyNetworksGeneralizedCliffordsSimulationStrategy,
)
from theboss.simulation_strategies.simulation_strategy_interface import (
    SimulationStrategyInterface,
)


def state_vector(state: SamplingState, instruction: Instruction, shots: int) -> Result:
    if not np.all(state.initial_state == 0):
        raise InvalidState("State vector is already set.")

    coefficient = instruction._all_params["coefficient"]
    occupation_numbers = instruction._all_params["occupation_numbers"]

    initial_state = coefficient * np.array(occupation_numbers)

    if not all_natural(initial_state):
        raise InvalidState(
            f"Invalid initial state specified: instruction={instruction}"
        )

    state.initial_state = np.rint(initial_state).astype(int)

    return Result(state=state)


def passive_linear(
    state: SamplingState, instruction: Instruction, shots: int
) -> Result:
    r"""Applies an interferometer to the circuit.

    This can be interpreted as placing another interferometer in the network, just
    before performing the sampling. This instruction is realized by multiplying
    current effective interferometer matrix with new interferometer matrix.

    Do note, that new interferometer matrix works as interferometer matrix on
    qumodes (provided as the arguments) and as an identity on every other mode.
    """
    _apply_matrix_on_modes(
        state=state,
        matrix=instruction._all_params["passive_block"],
        modes=instruction.modes,
    )

    return Result(state=state)


def _apply_matrix_on_modes(
    state: SamplingState, matrix: np.ndarray, modes: Tuple[int, ...]
) -> None:
    embedded = np.identity(len(state.interferometer), dtype=complex)

    embedded[np.ix_(modes, modes)] = matrix

    state.interferometer = embedded @ state.interferometer


def loss(state: SamplingState, instruction: Instruction, shots: int) -> Result:
    state.is_lossy = True

    _apply_matrix_on_modes(
        state=state,
        matrix=np.diag(instruction._all_params["transmissivity"]),
        modes=instruction.modes,
    )

    return Result(state=state)


def sampling(state: SamplingState, instruction: Instruction, shots: int) -> Result:
    initial_state = np.array(state.initial_state)
    permanent_calculator = RyserGuanPermanentCalculator(
        matrix=state.interferometer, input_state=initial_state
    )

    simulation_strategy = _get_sampling_simulation_strategy(state, permanent_calculator)

    sampling_simulator = BosonSamplingSimulator(simulation_strategy)

    samples = sampling_simulator.get_classical_simulation_results(
        initial_state, samples_number=shots
    )

    return Result(state=state, samples=list(map(tuple, samples)))


def _get_sampling_simulation_strategy(
    state: SamplingState, permanent_calculator: BSPermanentCalculatorInterface
) -> SimulationStrategyInterface:
    if not state.is_lossy:
        return GeneralizedCliffordsSimulationStrategy(permanent_calculator)

    _, singular_values, _ = np.linalg.svd(state.interferometer)

    if np.all(np.isclose(singular_values, singular_values[0])):
        return GeneralizedCliffordsUniformLossesSimulationStrategy(permanent_calculator)

    return LossyNetworksGeneralizedCliffordsSimulationStrategy(permanent_calculator)
